import MagicString from 'magic-string';
import { isExportNamedDeclaration, isExportDefaultDeclaration, isFunctionDeclaration, isVariableDeclaration, traverse, isReturnStatement, isTaggedTemplateExpression, isIdentifier, isNode, isCallExpression, isMemberExpression, isFunctionExpression, isArrowFunctionExpression, isImportDeclaration, isClassDeclaration, isTSEnumDeclaration, isTSTypeAliasDeclaration, isTSInterfaceDeclaration, isStringLiteral, isVariableDeclarator, isTSTypeAnnotation, isTSTypeLiteral, isTSPropertySignature, isTemplateLiteral, isObjectExpression, isImportSpecifier, isImportNamespaceSpecifier, isImportDefaultSpecifier, isExportDeclaration, isBlockStatement, isDeclaration, isObjectPattern, isArrayPattern, isExpressionStatement, isAssignmentExpression, isAwaitExpression } from '@babel/types';
import { transform, parserOptions, walkIdentifiers, parse as parse$1, compile } from '@vue/compiler-dom';
import { parseExpression, parse } from '@babel/parser';
import postcss from 'postcss';
import merge from 'merge-source-map';
import selectorParser from 'postcss-selector-parser';

const DEFINE_COMPONENT_HELPER = "defineComponent";
const USE_DEFAULTS_HELPER = "useDefaults";
const TO_REFS_HELPER = "toRefs";
const CSS_VARS_HELPER = "useCssVars";
const UN_REF_HELPER = "unref";
const BARE_CALL_MACROS = [
  "vineExpose",
  "vineOptions",
  "vineStyle",
  "vineStyle.scoped",
  "vineCustomElement"
];
const VINE_MACROS = [
  "vineProp",
  "vineProp.optional",
  "vineProp.withDefault",
  "vineEmits",
  ...BARE_CALL_MACROS
];
const SUPPORTED_CSS_LANGS = ["css", "scss", "sass", "less", "stylus", "postcss"];
const VUE_REACTIVITY_APIS = [
  "ref",
  "shallowRef",
  "computed",
  "reactive",
  "readonly",
  "watchEffect",
  "watchPostEffect",
  "watchSyncEffect",
  "watch",
  "triggerRef",
  "customRef",
  "shallowReactive",
  "shallowReadonly",
  "toRaw",
  "markRaw",
  "effectScope",
  "getCurrentScope",
  "onScopeDispose"
];
const TS_NODE_TYPES = [
  "TSAsExpression",
  "TSTypeAssertion",
  "TSNonNullExpression",
  "TSInstantiationExpression",
  "TSSatisfiesExpression"
];
const VineBindingTypes = {
  PROPS: "props",
  PROPS_ALIASED: "props-aliased",
  SETUP_LET: "setup-let",
  SETUP_CONST: "setup-const",
  SETUP_REACTIVE_CONST: "setup-reactive-const",
  SETUP_MAYBE_REF: "setup-maybe-ref",
  SETUP_REF: "setup-ref",
  LITERAL_CONST: "literal-const"
};

const vineRootScopeStatementTypeValidators = [
  isImportDeclaration,
  isExportNamedDeclaration,
  isFunctionDeclaration,
  isClassDeclaration,
  isTSEnumDeclaration,
  isVariableDeclaration,
  isTSTypeAliasDeclaration,
  isTSInterfaceDeclaration
];
function isVineCompFnDecl(target) {
  let isFound = false;
  if ((isExportNamedDeclaration(target) || isExportDefaultDeclaration(target)) && target.declaration) {
    target = target.declaration;
  }
  if (isFunctionDeclaration(target) || isVariableDeclaration(target)) {
    traverse(target, (node) => {
      if (!isFound && isReturnStatement(node) && node.argument && isVineTaggedTemplateString(node.argument)) {
        isFound = true;
      }
    });
  }
  return isFound;
}
function findVineCompFnDecls(root) {
  const vineFnComps = [];
  for (const stmt of root.program.body) {
    if (isVineCompFnDecl(stmt)) {
      vineFnComps.push(stmt);
    }
  }
  return vineFnComps;
}
function isDescendant(node, potentialDescendant) {
  const stack = [node];
  while (stack.length) {
    const currentNode = stack.pop();
    if (currentNode === potentialDescendant) {
      return true;
    }
    const children = Object.values(currentNode).filter((child) => Array.isArray(child) ? child.every(isNode) : isNode(child));
    for (const child of children) {
      if (Array.isArray(child)) {
        stack.push(...child);
      } else {
        stack.push(child);
      }
    }
  }
  return false;
}
function isVineTaggedTemplateString(node) {
  return isTaggedTemplateExpression(node) && isIdentifier(node.tag) && node.tag.name === "vine";
}
function isVineMacroCallExpression(node) {
  if (isCallExpression(node)) {
    const callee = node.callee;
    if (isIdentifier(callee)) {
      return VINE_MACROS.includes(callee.name);
    }
    if (isMemberExpression(callee)) {
      const obj = callee.object;
      const prop = callee.property;
      if (isIdentifier(obj) && isIdentifier(prop)) {
        return VINE_MACROS.includes(
          `${obj.name}.${prop.name}`
        );
      }
    }
  }
  return false;
}
function getVineMacroCalleeName(node) {
  const callee = node.callee;
  if (isIdentifier(callee)) {
    return callee.name;
  }
  if (isMemberExpression(callee)) {
    const obj = callee.object;
    const prop = callee.property;
    if (isIdentifier(obj) && isIdentifier(prop)) {
      return `${obj.name}.${prop.name}`;
    }
  }
  return "";
}
function isVineMacroOf(name) {
  return (node) => {
    if (isCallExpression(node)) {
      const callee = node.callee;
      if (isIdentifier(callee)) {
        return Array.isArray(name) ? name.includes(callee.name) : callee.name === name;
      }
      if (isMemberExpression(callee)) {
        const obj = callee.object;
        const prop = callee.property;
        if (isIdentifier(obj) && isIdentifier(prop)) {
          return Array.isArray(name) ? name.some((n) => `${obj.name}.${prop.name}`.includes(n)) : `${obj.name}.${prop.name}`.includes(name);
        }
        return false;
      }
    }
    return false;
  };
}
function isStatementContainsVineMacroCall(node) {
  let result = false;
  traverse(node, (descendant) => {
    if (isVineMacroCallExpression(descendant)) {
      result = true;
    }
  });
  return result;
}
function isVueReactivityApiCallExpression(node) {
  return isCallExpression(node) && isIdentifier(node.callee) && VUE_REACTIVITY_APIS.includes(node.callee.name);
}
function isValidVineRootScopeStatement(node) {
  return vineRootScopeStatementTypeValidators.some((check) => check(node));
}
function isTagTemplateStringContainsInterpolation(tagTmplNode) {
  return tagTmplNode.quasi.expressions.length > 0;
}
function getFunctionParams(fnItselfNode) {
  const params = [];
  if (isFunctionDeclaration(fnItselfNode)) {
    params.push(...fnItselfNode.params);
  } else if (isFunctionExpression(fnItselfNode) || isArrowFunctionExpression(fnItselfNode)) {
    params.push(...fnItselfNode.params);
  }
  return params;
}
function getFunctionInfo(fnDecl) {
  var _a, _b, _c, _d;
  let fnName = isFunctionDeclaration(fnDecl) ? (_b = (_a = fnDecl.id) == null ? void 0 : _a.name) != null ? _b : "" : "";
  let fnItselfNode = isFunctionDeclaration(fnDecl) ? fnDecl : void 0;
  let target = fnDecl;
  if (isExportNamedDeclaration(target) && target.declaration) {
    target = target.declaration;
  }
  if (isFunctionDeclaration(target)) {
    fnItselfNode = target;
    fnName = (_d = (_c = target.id) == null ? void 0 : _c.name) != null ? _d : "";
  } else if (isVariableDeclaration(target) && ((isFunctionExpression(target.declarations[0].init) || isArrowFunctionExpression(target.declarations[0].init)) && isIdentifier(target.declarations[0].id))) {
    fnItselfNode = target.declarations[0].init;
    fnName = target.declarations[0].id.name;
  }
  return {
    fnItselfNode,
    fnName
  };
}
function findVineTagTemplateStringReturn(node) {
  let templateReturn;
  let templateStringNode;
  traverse(node, (descendant) => {
    if (isReturnStatement(descendant)) {
      templateReturn = descendant;
      if (isVineTaggedTemplateString(descendant.argument)) {
        templateStringNode = descendant.argument;
      }
    }
  });
  return {
    templateReturn,
    templateStringNode
  };
}
function getImportStatments(root) {
  const importStmts = [];
  for (const stmt of root.program.body) {
    if (isImportDeclaration(stmt)) {
      importStmts.push(stmt);
    }
  }
  return importStmts;
}
function getTSTypeLiteralPropertySignatureName(tsTypeLit) {
  return isIdentifier(tsTypeLit.key) ? tsTypeLit.key.name : isStringLiteral(tsTypeLit.key) ? tsTypeLit.key.value : "";
}
function getAllVinePropMacroCall(fnItselfNode) {
  const allVinePropMacroCalls = [];
  traverse(fnItselfNode.body, {
    enter(node, parent) {
      if (!isVineMacroOf("vineProp")(node)) {
        return;
      }
      const propVarDeclarator = parent.find((ancestor) => isVariableDeclarator(ancestor.node) && isIdentifier(ancestor.node.id));
      if (!propVarDeclarator) {
        return;
      }
      const propVarIdentifier = propVarDeclarator.node.id;
      allVinePropMacroCalls.push([node, propVarIdentifier]);
    }
  });
  return allVinePropMacroCalls;
}
function unwrapTSNode(node) {
  if (TS_NODE_TYPES.includes(node.type)) {
    return unwrapTSNode(node.expression);
  } else {
    return node;
  }
}
function isStaticNode(node) {
  node = unwrapTSNode(node);
  switch (node.type) {
    case "UnaryExpression":
      return isStaticNode(node.argument);
    case "LogicalExpression":
    case "BinaryExpression":
      return isStaticNode(node.left) && isStaticNode(node.right);
    case "ConditionalExpression": {
      return isStaticNode(node.test) && isStaticNode(node.consequent) && isStaticNode(node.alternate);
    }
    case "SequenceExpression":
    case "TemplateLiteral":
      return node.expressions.every((expr) => isStaticNode(expr));
    case "ParenthesizedExpression":
      return isStaticNode(node.expression);
    case "StringLiteral":
    case "NumericLiteral":
    case "BooleanLiteral":
    case "NullLiteral":
    case "BigIntLiteral":
      return true;
  }
  return false;
}
function isCallOf(node, test) {
  return !!(node && test && node.type === "CallExpression" && node.callee.type === "Identifier" && (typeof test === "string" ? node.callee.name === test : test(node.callee.name)));
}
function isLiteralNode(node) {
  return node.type.endsWith("Literal");
}
function canNeverBeRef(node, userReactiveImport) {
  if (isCallOf(node, userReactiveImport)) {
    return true;
  }
  switch (node.type) {
    case "UnaryExpression":
    case "BinaryExpression":
    case "ArrayExpression":
    case "ObjectExpression":
    case "FunctionExpression":
    case "ArrowFunctionExpression":
    case "UpdateExpression":
    case "ClassExpression":
    case "TaggedTemplateExpression":
      return true;
    case "SequenceExpression":
      return canNeverBeRef(
        node.expressions[node.expressions.length - 1],
        userReactiveImport
      );
    default:
      if (isLiteralNode(node)) {
        return true;
      }
      return false;
  }
}

const colorMap = {
  bold: "1",
  italic: "3",
  underline: "4",
  inverse: "7",
  strikethrough: "9",
  white: "37",
  grey: "90",
  black: "30",
  blue: "34",
  cyan: "36",
  green: "32",
  magenta: "35",
  red: "31",
  yellow: "33",
  bgWhite: "47",
  bgGrey: "49",
  bgBlack: "40",
  bgBlue: "44",
  bgCyan: "46",
  bgGreen: "42",
  bgMagenta: "45",
  bgRed: "41",
  bgYellow: "43"
};
function colorful(str, options = []) {
  const colors = options.map((option) => {
    return colorMap[option];
  }).join(";");
  return `\x1B[${colors}m${str}\x1B[0m`;
}
function createColorful(options) {
  return (str) => colorful(str, options);
}
Object.keys(colorMap).reduce((acc, key) => {
  acc[key] = createColorful([key]);
  return acc;
}, {});

const whiteFgRedBg = createColorful(["white", "bgRed"]);
const whiteFgYellowBg = createColorful(["white", "bgYellow"]);
const redFgBold = createColorful(["red", "bold"]);
const ErrorLabel = whiteFgRedBg(" Error ");
const WarningLabel = whiteFgYellowBg(" Warning ");
function makeDiagnostic(label, { msg, fileId, location, msgColorful }) {
  const pos = location == null ? void 0 : location.start;
  const posString = pos ? `${pos.line}:${pos.column}` : "";
  return `
${label} File: ${fileId} ${posString}
${msgColorful(msg)}
`;
}
function vineErr(vineFileCtx, { msg, location }) {
  const full = makeDiagnostic(ErrorLabel, {
    msg,
    fileId: vineFileCtx.fileId,
    location,
    msgColorful: redFgBold
  });
  return { full, msg, location };
}
function vineWarn(vineFileCtx, { msg, location }) {
  const full = makeDiagnostic(WarningLabel, {
    msg,
    fileId: vineFileCtx.fileId,
    location,
    msgColorful: whiteFgYellowBg
  });
  return { full, msg, location };
}

function wrapVineValidatorWithLog(validators) {
  return process.env.VINE_DEV_VITEST === "true" ? validators.map((validator) => (...args) => {
    const isPass = validator(...args);
    console.log(`${colorful(" VINE VALIDATE ", ["bgGreen", "white"])} ${validator.name} => ${colorful(isPass ? "PASS" : "FAIL", [isPass ? "green" : "red"])}`);
    return isPass;
  }) : validators;
}
function logMacroAssert(macroName, isPass) {
  if (process.env.VINE_DEV_VITEST !== "true") {
    return;
  }
  console.log(`${colorful(" VINE MACRO ASSERT ", ["bgMagenta", "white"])} ${macroName} => ${colorful(isPass ? "PASS" : "FAIL", [isPass ? "green" : "red"])}`);
}
function validateNoOutsideMacroCalls({ vineCompilerHooks, vineFileCtx, vineCompFns }, root) {
  let isMacroInsideVineCompFn = true;
  traverse(root, (node) => {
    if (!isVineMacroCallExpression(node)) {
      return;
    }
    isMacroInsideVineCompFn = vineCompFns.some(
      (vineCompFn) => isDescendant(vineCompFn, node)
    );
    if (!isMacroInsideVineCompFn) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Vine macro calls must be inside Vue Vine component function!",
            location: node.loc
          }
        )
      );
    }
  });
  return isMacroInsideVineCompFn;
}
function validateNoInvalidTypesRootScopeStmt({ vineCompilerHooks, vineFileCtx }, root) {
  for (const stmt of root.program.body) {
    if (isVariableDeclaration(stmt)) {
      let hasVueReactivityApiCallInRootVarDeclaration = false;
      traverse(stmt, (node) => {
        if (isVueReactivityApiCallExpression(node)) {
          vineCompilerHooks.onError(
            vineErr(
              vineFileCtx,
              {
                msg: "Vue API calls are not allowed to be called in Vine root scope!",
                location: node.loc
              }
            )
          );
          hasVueReactivityApiCallInRootVarDeclaration = true;
        }
      });
      if (hasVueReactivityApiCallInRootVarDeclaration) {
        return false;
      }
    } else if (!isValidVineRootScopeStatement(stmt)) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Invalid root scope statements must be inside Vue Vine component function!",
            location: stmt.loc
          }
        )
      );
      return false;
    }
  }
  return true;
}
function validateVineTemplateStringUsage({ vineCompilerHooks, vineFileCtx }, vineCompFn) {
  let isVineTemplateUsagePass = true;
  let vineTemplateStrCount = 0;
  traverse(vineCompFn, (node) => {
    if (!isVineTaggedTemplateString(node)) {
      return;
    }
    vineTemplateStrCount += 1;
    if (vineTemplateStrCount > 1) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Multiple vine tagged template are not allowed inside Vine component function",
            location: node.loc
          }
        )
      );
      isVineTemplateUsagePass = false;
    }
    if (isTagTemplateStringContainsInterpolation(node)) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Vine template string are not allowed to contain interpolation!",
            location: node.loc
          }
        )
      );
      isVineTemplateUsagePass = false;
    }
  });
  return isVineTemplateUsagePass;
}
function assertVineStyleUsage({ vineCompilerHooks, vineFileCtx }, vineStyleMacroCallNode) {
  const vineStyleArgsLength = vineStyleMacroCallNode.arguments.length;
  const macroCallee = vineStyleMacroCallNode.callee;
  if (vineStyleArgsLength !== 1) {
    vineCompilerHooks.onError(
      vineErr(
        vineFileCtx,
        {
          msg: `\`vineStyle\` ${vineStyleArgsLength > 1 ? "can only" : "must"} have one string argument'`,
          location: macroCallee.loc
        }
      )
    );
    return false;
  }
  const theOnlyArg = vineStyleMacroCallNode.arguments[0];
  if (isStringLiteral(theOnlyArg)) ; else if (isTemplateLiteral(theOnlyArg)) {
    if (theOnlyArg.expressions.length > 0) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "vineStyle argument must be a plain string, template string interpolation is not allowed",
            location: theOnlyArg.loc
          }
        )
      );
      return false;
    }
  } else if (isTaggedTemplateExpression(theOnlyArg)) {
    if (theOnlyArg.tag && isIdentifier(theOnlyArg.tag) && !SUPPORTED_CSS_LANGS.includes(theOnlyArg.tag.name)) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "vineStyle CSS language only supports: `css`, `scss`, `sass`, `less`, `stylus` and `postcss`",
            location: theOnlyArg.tag.loc
          }
        )
      );
      return false;
    }
  } else {
    vineCompilerHooks.onError(
      vineErr(
        vineFileCtx,
        {
          msg: "vineStyle's argument can only be string",
          location: theOnlyArg.loc
        }
      )
    );
    return false;
  }
  return true;
}
function assertCanOnlyHaveOneObjLiteralArg({ vineCompilerHooks, vineFileCtx }, macroCallNode) {
  const macroCallArgsLength = macroCallNode.arguments.length;
  const macroCallee = macroCallNode.callee;
  const errMsg = `\`${macroCallee.name}\` ${macroCallArgsLength > 1 ? "can only" : "must"} have one object literal argument`;
  if (macroCallArgsLength === 1) {
    const theOnlyArg = macroCallNode.arguments[0];
    if (!isObjectExpression(theOnlyArg)) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: errMsg,
            location: theOnlyArg.loc
          }
        )
      );
      return false;
    }
  } else {
    vineCompilerHooks.onError(
      vineErr(
        vineFileCtx,
        {
          msg: errMsg,
          location: macroCallee.loc
        }
      )
    );
    return false;
  }
  return true;
}
function assertVineEmitsUsage({ vineCompilerHooks, vineFileCtx }, macroCallNode, parent) {
  var _a;
  const macroCallee = macroCallNode.callee;
  const typeParams = (_a = macroCallNode.typeParameters) == null ? void 0 : _a.params;
  const typeParamsLength = typeParams == null ? void 0 : typeParams.length;
  const errMsg = `\`${macroCallee.name}\` ${typeParamsLength && typeParamsLength > 1 ? "can only" : "must"} have one type parameter!`;
  let isVineEmitsTypeParamCorrect = true;
  if (typeParamsLength !== 1) {
    vineCompilerHooks.onError(
      vineErr(
        vineFileCtx,
        {
          msg: errMsg,
          location: macroCallee.loc
        }
      )
    );
    isVineEmitsTypeParamCorrect = false;
  } else {
    const theOnlyTypeParam = typeParams == null ? void 0 : typeParams[0];
    if (!isTSTypeLiteral(theOnlyTypeParam)) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Vue Vine component function's vineEmits type must be object literal!",
            location: theOnlyTypeParam == null ? void 0 : theOnlyTypeParam.loc
          }
        )
      );
      isVineEmitsTypeParamCorrect = false;
    }
    const properties = theOnlyTypeParam == null ? void 0 : theOnlyTypeParam.members;
    if (!properties.every((prop) => isTSPropertySignature(prop))) {
      isVineEmitsTypeParamCorrect = false;
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Vue Vine component function's vineEmits type must be object literal! And all properties' key must be string literal or identifier",
            location: theOnlyTypeParam == null ? void 0 : theOnlyTypeParam.loc
          }
        )
      );
    }
  }
  const varDeclaratorThatMayBeInside = parent == null ? void 0 : parent.find((ancestor) => isVariableDeclarator(ancestor.node));
  if (varDeclaratorThatMayBeInside) {
    const varDeclThatMayBeInsideNode = varDeclaratorThatMayBeInside.node;
    if (!isIdentifier(varDeclThatMayBeInsideNode.id)) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "the declaration of `vineEmits` macro call must be an identifier",
            location: varDeclThatMayBeInsideNode.id.loc
          }
        )
      );
      isVineEmitsTypeParamCorrect = false;
    }
  }
  return isVineEmitsTypeParamCorrect;
}
function validateMacrosUsage(validatorCtx, vineCompFn) {
  const { vineCompilerHooks, vineFileCtx } = validatorCtx;
  const macroCountMap = {
    vineStyle: {
      count: 0,
      asserts: [
        assertVineStyleUsage
      ]
    },
    vineEmits: {
      count: 0,
      asserts: [
        assertVineEmitsUsage
      ]
    },
    vineExpose: {
      count: 0,
      asserts: [
        assertCanOnlyHaveOneObjLiteralArg
      ]
    },
    vineOptions: {
      count: 0,
      asserts: [
        assertCanOnlyHaveOneObjLiteralArg
      ]
    },
    vineCustomElement: {
      count: 0,
      asserts: []
    }
  };
  let isCountCorrect = true;
  let isBareCallMacrosPass = true;
  traverse(vineCompFn, {
    enter(node, parent) {
      if (!isVineMacroCallExpression(node)) {
        return;
      }
      const macroCalleeName = getVineMacroCalleeName(node);
      if (!macroCalleeName) {
        return;
      }
      const macroName = macroCalleeName.includes(".") ? macroCalleeName.split(".")[0] : macroCalleeName;
      if (macroCountMap[macroName]) {
        macroCountMap[macroName].count += 1;
        macroCountMap[macroName].node = node;
        macroCountMap[macroName].parent = [...parent];
      }
      const VarDeclThatMacroBeInside = parent.find((ancestor) => isVariableDeclaration(ancestor.node));
      if (VarDeclThatMacroBeInside && BARE_CALL_MACROS.includes(macroName)) {
        isBareCallMacrosPass = false;
        vineCompilerHooks.onError(
          vineErr(
            vineFileCtx,
            {
              msg: `\`${macroName}\` macro call is not allowed to be inside a variable declaration`,
              location: VarDeclThatMacroBeInside == null ? void 0 : VarDeclThatMacroBeInside.node.loc
            }
          )
        );
      }
    }
  });
  const macroMoreThanOnce = Object.entries(macroCountMap).filter(([, it]) => it.count > 1);
  if (macroMoreThanOnce.length > 0) {
    macroMoreThanOnce.forEach(([macroName, it]) => {
      var _a;
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: `Multiple \`${macroName}\` calls are not allowed inside Vine component function`,
            location: (_a = it.node) == null ? void 0 : _a.loc
          }
        )
      );
    });
    isCountCorrect = false;
  }
  const isAssertsPass = Object.entries(macroCountMap).map(
    ([macroName, it]) => it.asserts.map((assert) => {
      const isPass = it.node ? assert(validatorCtx, it.node, it.parent) : true;
      logMacroAssert(macroName, isPass);
      return isPass;
    }).every(Boolean)
  ).every(Boolean);
  return isCountCorrect && isBareCallMacrosPass && isAssertsPass;
}
function validateVineFunctionCompProps({ vineCompilerHooks, vineFileCtx }, vineCompFnDecl) {
  const { fnItselfNode } = getFunctionInfo(vineCompFnDecl);
  const vineCompFnParams = fnItselfNode ? getFunctionParams(fnItselfNode) : [];
  const vineCompFnParamsLength = vineCompFnParams.length;
  let vinePropMacroCallCount = 0;
  const isCheckVinePropMacroCallPass = () => {
    let isVinePropCheckPass = true;
    traverse(vineCompFnDecl, {
      enter(node, parent) {
        var _a;
        if (!isVineMacroOf("vineProp")(node)) {
          return;
        }
        vinePropMacroCallCount += 1;
        const macroCalleeName = getVineMacroCalleeName(node);
        if (!macroCalleeName) {
          return;
        }
        if (macroCalleeName !== "vineProp.withDefault") {
          const typeParamsLength = (_a = node.typeParameters) == null ? void 0 : _a.params.length;
          if (typeParamsLength !== 1) {
            vineCompilerHooks.onError(
              vineErr(
                vineFileCtx,
                {
                  msg: `\`${macroCalleeName}\` macro call ${typeParamsLength && typeParamsLength > 1 ? "can only" : "must"} have a type parameter to specify the prop's type`,
                  location: node.loc
                }
              )
            );
            isVinePropCheckPass = false;
          }
        } else if (!node.arguments.length) {
          isVinePropCheckPass = false;
          vineCompilerHooks.onError(
            vineErr(
              vineFileCtx,
              {
                msg: `\`${macroCalleeName}\` macro call must have at least 1 argument`,
                location: node.loc
              }
            )
          );
        }
        const parentVarDecl = parent.find((ancestor) => isVariableDeclaration(ancestor.node));
        const parentVarDeclarator = parent.find((ancestor) => isVariableDeclarator(ancestor.node));
        if (!parentVarDecl || !parentVarDeclarator) {
          isVinePropCheckPass = false;
          vineCompilerHooks.onError(
            vineErr(
              vineFileCtx,
              {
                msg: "`vineProp` macro call must be inside a `const` variable declaration",
                location: node.loc
              }
            )
          );
        } else {
          const isConst = parentVarDecl.node.kind === "const";
          if (!isConst) {
            isVinePropCheckPass = false;
            vineCompilerHooks.onError(
              vineErr(
                vineFileCtx,
                {
                  msg: "`vineProp` macro call must be inside a `const` declaration",
                  location: parentVarDecl.node.loc
                }
              )
            );
          }
          const varDeclarator = parentVarDeclarator.node;
          if (!isIdentifier(varDeclarator.id)) {
            isVinePropCheckPass = false;
            vineCompilerHooks.onError(
              vineErr(
                vineFileCtx,
                {
                  msg: "the declaration of `vineProp` macro call must be an identifier",
                  location: varDeclarator.id.loc
                }
              )
            );
          }
        }
      }
    });
    return isVinePropCheckPass;
  };
  if (vineCompFnParamsLength === 0) {
    return isCheckVinePropMacroCallPass();
  } else if (vineCompFnParamsLength === 1) {
    const theOnlyFormalParam = vineCompFnParams[0];
    let isCheckFormalParamsPropPass = true;
    if (!isIdentifier(theOnlyFormalParam)) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "If you're defining a Vine component function's props with formal parameter, it must be one and only identifier",
            location: theOnlyFormalParam.loc
          }
        )
      );
      isCheckFormalParamsPropPass = false;
    }
    const theOnlyFormalParamTypeAnnotation = theOnlyFormalParam.typeAnnotation;
    if (!theOnlyFormalParamTypeAnnotation) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Vine component function's props must have a type annotation",
            location: theOnlyFormalParam.loc
          }
        )
      );
      isCheckFormalParamsPropPass = false;
    }
    if (isTSTypeAnnotation(theOnlyFormalParamTypeAnnotation) && isTSTypeLiteral(theOnlyFormalParamTypeAnnotation.typeAnnotation)) {
      const propsTypeAnnotation = theOnlyFormalParamTypeAnnotation.typeAnnotation;
      const isTSTypeAnnotationValid = propsTypeAnnotation.members.every(
        (member) => isTSPropertySignature(member) && (isStringLiteral(member.key) || isIdentifier(member.key))
      );
      if (!isTSTypeAnnotationValid) {
        vineCompilerHooks.onError(
          vineErr(
            vineFileCtx,
            {
              msg: "Vine component function's props type annotation must be an object literal, only contains properties signature, and all properties' key must be string literal or identifier",
              location: propsTypeAnnotation.loc
            }
          )
        );
        isCheckFormalParamsPropPass = false;
      }
    } else {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Vine component function's props type annotation must be an object literal",
            location: theOnlyFormalParamTypeAnnotation == null ? void 0 : theOnlyFormalParamTypeAnnotation.loc
          }
        )
      );
      isCheckFormalParamsPropPass = false;
    }
    if (vinePropMacroCallCount > 0) {
      vineCompilerHooks.onError(
        vineErr(
          vineFileCtx,
          {
            msg: "Vine component function's props can only be defined with formal parameter or `vineProp` macro calls, not both",
            location: theOnlyFormalParam.loc
          }
        )
      );
      isCheckFormalParamsPropPass = false;
    }
    if (!isCheckFormalParamsPropPass) {
      isCheckVinePropMacroCallPass();
      return false;
    }
    return isCheckVinePropMacroCallPass();
  }
  vineCompilerHooks.onError(
    vineErr(
      vineFileCtx,
      {
        msg: "Vine component function can only have one parameter",
        location: vineCompFnDecl.loc
      }
    )
  );
  return false;
}
const validatesFromRoot = wrapVineValidatorWithLog([
  validateNoOutsideMacroCalls,
  validateNoInvalidTypesRootScopeStmt
]);
const validatesFromVineFn = wrapVineValidatorWithLog([
  validateVineTemplateStringUsage,
  validateMacrosUsage,
  validateVineFunctionCompProps
]);
function createVineValidator(vineCompilerHooks, vineFileCtx, vineCompFns) {
  return (validates, fromNode) => {
    return validates.map(
      (validateFn) => validateFn(
        {
          vineCompilerHooks,
          vineFileCtx,
          vineCompFns
        },
        fromNode
      )
    ).filter(Boolean).length < validates.length;
  };
}
function validateVine(compilerHooks, vineFileCtx, vineCompFnDecls) {
  const findValidateError = createVineValidator(compilerHooks, vineFileCtx, vineCompFnDecls);
  if (findValidateError(validatesFromRoot, vineFileCtx.root)) {
    return false;
  }
  let hasErrInVineCompFns = false;
  for (const vineCompFnDecl of vineCompFnDecls) {
    hasErrInVineCompFns = findValidateError(validatesFromVineFn, vineCompFnDecl);
  }
  return !hasErrInVineCompFns;
}

function getDefaultExportFromCjs (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

function pad (hash, len) {
  while (hash.length < len) {
    hash = '0' + hash;
  }
  return hash;
}

function fold (hash, text) {
  var i;
  var chr;
  var len;
  if (text.length === 0) {
    return hash;
  }
  for (i = 0, len = text.length; i < len; i++) {
    chr = text.charCodeAt(i);
    hash = ((hash << 5) - hash) + chr;
    hash |= 0;
  }
  return hash < 0 ? hash * -2 : hash;
}

function foldObject (hash, o, seen) {
  return Object.keys(o).sort().reduce(foldKey, hash);
  function foldKey (hash, key) {
    return foldValue(hash, o[key], key, seen);
  }
}

function foldValue (input, value, key, seen) {
  var hash = fold(fold(fold(input, key), toString$1(value)), typeof value);
  if (value === null) {
    return fold(hash, 'null');
  }
  if (value === undefined) {
    return fold(hash, 'undefined');
  }
  if (typeof value === 'object' || typeof value === 'function') {
    if (seen.indexOf(value) !== -1) {
      return fold(hash, '[Circular]' + key);
    }
    seen.push(value);

    var objHash = foldObject(hash, value, seen);

    if (!('valueOf' in value) || typeof value.valueOf !== 'function') {
      return objHash;
    }

    try {
      return fold(objHash, String(value.valueOf()))
    } catch (err) {
      return fold(objHash, '[valueOf exception]' + (err.stack || err.message))
    }
  }
  return fold(hash, value.toString());
}

function toString$1 (o) {
  return Object.prototype.toString.call(o);
}

function sum (o) {
  return pad(foldValue(0, o, '', []).toString(16), 8);
}

var hashSum = sum;

var hashId = /*@__PURE__*/getDefaultExportFromCjs(hashSum);

function lexBinding(content, start) {
  let state = 0 /* inParens */;
  let parenDepth = 0;
  for (let i = start; i < content.length; i++) {
    const char = content.charAt(i);
    switch (state) {
      case 0 /* inParens */:
        if (char === "'") {
          state = 1 /* inSingleQuoteString */;
        } else if (char === '"') {
          state = 2 /* inDoubleQuoteString */;
        } else if (char === "(") {
          parenDepth++;
        } else if (char === ")") {
          if (parenDepth > 0) {
            parenDepth--;
          } else {
            return i;
          }
        }
        break;
      case 1 /* inSingleQuoteString */:
        if (char === "'") {
          state = 0 /* inParens */;
        }
        break;
      case 2 /* inDoubleQuoteString */:
        if (char === '"') {
          state = 0 /* inParens */;
        }
        break;
    }
  }
  return null;
}
function normalizeExpression(exp) {
  exp = exp.trim();
  if (exp[0] === "'" && exp[exp.length - 1] === "'" || exp[0] === '"' && exp[exp.length - 1] === '"') {
    return exp.slice(1, -1);
  }
  return exp;
}
const vBindRE = /v-bind\s*\(/g;
function parseCssVars(styles, hook) {
  const vars = [];
  styles.forEach((style) => {
    let match = null;
    const content = style.replace(/\/\*([\s\S]*?)\*\//g, "");
    for (match = vBindRE.exec(content); match; match = vBindRE.exec(content)) {
      const start = match.index + match[0].length;
      const end = lexBinding(content, start);
      if (end !== null) {
        hook && hook.getIndex(start, end);
        const variable = normalizeExpression(content.slice(start, end));
        if (!vars.includes(variable)) {
          vars.push(variable);
        }
      }
    }
  });
  return vars;
}

function makeMap(str, expectsLowerCase) {
  const map = /* @__PURE__ */ Object.create(null);
  const list = str.split(",");
  for (let i = 0; i < list.length; i++) {
    map[list[i]] = true;
  }
  return expectsLowerCase ? (val) => !!map[val.toLowerCase()] : (val) => !!map[val];
}

const camelizeRE = /-(\w)/g;
function cacheStringFunction(fn) {
  const cache = /* @__PURE__ */ Object.create(null);
  return (str) => {
    const hit = cache[str];
    return hit || (cache[str] = fn(str));
  };
}
function showIf(condition, s, not) {
  return condition ? s : not != null ? not : "";
}
function filterJoin(arr, join) {
  return arr.filter(Boolean).join(join);
}
function getStyleLangFileExt(styleLang) {
  if (styleLang === "postcss") {
    return "css";
  } else if (styleLang === "stylus") {
    return "styl";
  }
  return styleLang;
}
function appendToMapArray(storeMap, key, value) {
  const arr = storeMap.get(key);
  if (!arr) {
    storeMap.set(key, [value]);
  } else {
    arr.push(value);
  }
}
const camelize = cacheStringFunction((str) => {
  return str.replace(camelizeRE, (_, c) => c ? c.toUpperCase() : "");
});
const capitalize = cacheStringFunction(
  (str) => str.charAt(0).toUpperCase() + str.slice(1)
);
const isBuiltInDirective = /* @__PURE__ */ makeMap(
  "bind,cloak,else-if,else,for,html,if,model,on,once,pre,show,slot,text,memo"
);

const templateUsageCheckCache = /* @__PURE__ */ new Map();
const forAliasRE = /([\s\S]*?)\s+(?:in|of)\s+([\s\S]*)/;
function resolveTemplateUsageCheckString(content, astRoot) {
  const cached = templateUsageCheckCache.get(content);
  if (cached) {
    return cached;
  }
  let code = "";
  transform(astRoot, {
    nodeTransforms: [
      (node) => {
        if (node.type === 1) {
          if (!parserOptions.isNativeTag(node.tag) && !parserOptions.isBuiltInComponent(node.tag)) {
            code += `,${camelize(node.tag)},${capitalize(camelize(node.tag))}`;
          }
          for (let i = 0; i < node.props.length; i++) {
            const prop = node.props[i];
            if (prop.type === 7) {
              if (!isBuiltInDirective(prop.name)) {
                code += `,v${capitalize(camelize(prop.name))}`;
              }
              if (prop.exp) {
                code += `,${processExp(
                  prop.exp.content,
                  prop.name
                )}`;
              }
            }
          }
        } else if (node.type === 5) {
          code += `,${processExp(
            node.content.content
          )}`;
        }
      }
    ]
  });
  code += ";";
  templateUsageCheckCache.set(content, code);
  return code;
}
function processExp(exp, dir) {
  if (/ as\s+\w|<.*>|:/.test(exp)) {
    if (dir === "slot") {
      exp = `(${exp})=>{}`;
    } else if (dir === "on") {
      exp = `()=>{return ${exp}}`;
    } else if (dir === "for") {
      const inMatch = exp.match(forAliasRE);
      if (inMatch) {
        let [, LHS, RHS] = inMatch;
        LHS = LHS.trim().replace(/^\(|\)$/g, "");
        return processExp(`(${LHS})=>{}`) + processExp(RHS);
      }
    }
    let ret = "";
    const ast = parseExpression(exp, { plugins: ["typescript"] });
    walkIdentifiers(ast, (node) => {
      ret += `,${node.name}`;
    });
    return ret;
  }
  return stripStrings(exp);
}
function stripStrings(exp) {
  return exp.replace(/'[^']*'|"[^"]*"/g, "").replace(/`[^`]+`/g, stripTemplateString);
}
function stripTemplateString(str) {
  const interpMatch = str.match(/\${[^}]+}/g);
  if (interpMatch) {
    return interpMatch.map((m) => m.slice(2, -1)).join(",");
  }
  return "";
}
function isImportUsed(vineCompFnCtx, local) {
  return !!vineCompFnCtx.templateAst && new RegExp(
    `[^\\w$_]${local.replace(/\$/g, "\\$")}[^\\w$_]`
  ).test(
    resolveTemplateUsageCheckString(
      vineCompFnCtx.templateSource,
      vineCompFnCtx.templateAst
    )
  );
}

var __defProp$5 = Object.defineProperty;
var __getOwnPropSymbols$5 = Object.getOwnPropertySymbols;
var __hasOwnProp$5 = Object.prototype.hasOwnProperty;
var __propIsEnum$5 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$5 = (obj, key, value) => key in obj ? __defProp$5(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$5 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$5.call(b, prop))
      __defNormalProp$5(a, prop, b[prop]);
  if (__getOwnPropSymbols$5)
    for (var prop of __getOwnPropSymbols$5(b)) {
      if (__propIsEnum$5.call(b, prop))
        __defNormalProp$5(a, prop, b[prop]);
    }
  return a;
};
function storeTheOnlyMacroCallArg(macroName, callback) {
  const findMacroCallNode = (fnItselfNode) => {
    let vineExposeMacroCall;
    traverse(fnItselfNode, {
      enter(descendant) {
        if (isVineMacroOf(macroName)(descendant)) {
          vineExposeMacroCall = descendant;
        }
      }
    });
    return vineExposeMacroCall;
  };
  return (analyzeCtx, fnItselfNode) => {
    const macroCall = findMacroCallNode(fnItselfNode);
    if (!macroCall) {
      return;
    }
    const macroCallArg = macroCall.arguments[0];
    if (!macroCallArg) {
      return;
    }
    callback(analyzeCtx, macroCallArg);
  };
}
const analyzeVineExpose = storeTheOnlyMacroCallArg(
  "vineExpose",
  ({ vineCompFnCtx }, macroCallArg) => {
    vineCompFnCtx.expose = macroCallArg;
  }
);
const analyzeVineOptions = storeTheOnlyMacroCallArg(
  "vineOptions",
  ({ vineCompFnCtx }, macroCallArg) => {
    vineCompFnCtx.options = macroCallArg;
  }
);
function registerBinding(bindings, node, type) {
  bindings[node.name] = type;
}
function walkPattern(node, bindings, isConst) {
  if (node.type === "Identifier") {
    const type = isConst ? VineBindingTypes.SETUP_MAYBE_REF : VineBindingTypes.SETUP_LET;
    registerBinding(bindings, node, type);
  } else if (node.type === "RestElement") {
    const type = isConst ? VineBindingTypes.SETUP_CONST : VineBindingTypes.SETUP_LET;
    registerBinding(bindings, node.argument, type);
  } else if (node.type === "ObjectPattern") {
    walkObjectPattern(node, bindings, isConst);
  } else if (node.type === "ArrayPattern") {
    walkArrayPattern(node, bindings, isConst);
  } else if (node.type === "AssignmentPattern") {
    if (node.left.type === "Identifier") {
      const type = isConst ? VineBindingTypes.SETUP_MAYBE_REF : VineBindingTypes.SETUP_LET;
      registerBinding(bindings, node.left, type);
    } else {
      walkPattern(node.left, bindings, isConst);
    }
  }
}
function walkArrayPattern(node, bindings, isConst) {
  for (const e of node.elements) {
    e && walkPattern(e, bindings, isConst);
  }
}
function walkObjectPattern(node, bindings, isConst) {
  for (const p of node.properties) {
    if (p.type === "ObjectProperty") {
      if (p.key.type === "Identifier" && p.key === p.value) {
        const type = isConst ? VineBindingTypes.SETUP_MAYBE_REF : VineBindingTypes.SETUP_LET;
        registerBinding(bindings, p.key, type);
      } else {
        walkPattern(p.value, bindings, isConst);
      }
    } else {
      const type = isConst ? VineBindingTypes.SETUP_CONST : VineBindingTypes.SETUP_LET;
      registerBinding(bindings, p.argument, type);
    }
  }
}
function analyzeVariableDeclarationForBindings({ vineFileCtx, vineCompFnCtx }, stmt) {
  const userImportAliases = __spreadValues$5({}, vineFileCtx.vueImportAliases);
  const userReactiveBinding = userImportAliases.reactive;
  const allDeclarators = stmt.declarations;
  const isConst = stmt.kind === "const";
  const isAllLiteral = isConst && allDeclarators.every(
    (decl) => isIdentifier(decl.id) && decl.init && isStaticNode(decl.init)
  );
  for (const varDecl of allDeclarators) {
    if (isIdentifier(varDecl.id) && varDecl.init) {
      let bindingType;
      if (isAllLiteral || isConst && isStaticNode(varDecl.init)) {
        bindingType = VineBindingTypes.LITERAL_CONST;
      } else if (isCallOf(varDecl.init, userReactiveBinding)) {
        bindingType = isConst ? VineBindingTypes.SETUP_REACTIVE_CONST : VineBindingTypes.SETUP_LET;
      } else if (isConst && canNeverBeRef(varDecl.init, userReactiveBinding)) {
        bindingType = VineBindingTypes.SETUP_CONST;
      } else if (isConst) {
        if (isCallOf(
          varDecl.init,
          (m) => m === userImportAliases.ref || m === userImportAliases.computed || m === userImportAliases.shallowRef || m === userImportAliases.customRef || m === userImportAliases.toRef
        )) {
          bindingType = VineBindingTypes.SETUP_REF;
        } else {
          bindingType = VineBindingTypes.SETUP_MAYBE_REF;
        }
      } else {
        bindingType = VineBindingTypes.SETUP_LET;
      }
      registerBinding(vineCompFnCtx.bindings, varDecl.id, bindingType);
    } else if (isObjectPattern(varDecl.id)) {
      walkObjectPattern(
        varDecl.id,
        vineCompFnCtx.bindings,
        isConst
      );
    } else if (isArrayPattern(varDecl.id)) {
      walkArrayPattern(
        varDecl.id,
        vineCompFnCtx.bindings,
        isConst
      );
    }
  }
  return isAllLiteral;
}
function analyzeVineFnBodyStmtForBindings(analyzeCtx, stmt) {
  const { vineCompFnCtx } = analyzeCtx;
  let isAllLiteral = false;
  switch (stmt.type) {
    case "VariableDeclaration":
      isAllLiteral = analyzeVariableDeclarationForBindings(analyzeCtx, stmt);
      break;
    case "TSEnumDeclaration":
      isAllLiteral = stmt.members.every(
        (member) => !member.initializer || isStaticNode(member.initializer)
      );
      vineCompFnCtx.bindings[stmt.id.name] = isAllLiteral ? VineBindingTypes.LITERAL_CONST : VineBindingTypes.SETUP_CONST;
      break;
    case "FunctionDeclaration":
    case "ClassDeclaration":
      vineCompFnCtx.bindings[stmt.id.name] = VineBindingTypes.SETUP_CONST;
      break;
  }
  return isAllLiteral;
}
function getVineStyleSource(vineStyleArg) {
  let styleLang = "css";
  let styleSource = "";
  let range;
  if (isTaggedTemplateExpression(vineStyleArg)) {
    const { tag } = vineStyleArg;
    if (isIdentifier(tag)) {
      styleLang = tag.name;
      const styleSourceNode = vineStyleArg.quasi.quasis[0];
      styleSource = styleSourceNode.value.raw;
      range = [styleSourceNode.start, styleSourceNode.end];
    }
  } else if (isStringLiteral(vineStyleArg)) {
    styleSource = vineStyleArg.value;
    range = [vineStyleArg.start, vineStyleArg.end];
  } else if (isTemplateLiteral(vineStyleArg)) {
    const styleSourceNode = vineStyleArg.quasis[0];
    styleSource = styleSourceNode.value.raw;
    range = [styleSourceNode.start, styleSourceNode.end];
  }
  return {
    styleSource,
    styleLang,
    range
  };
}
const analyzeVineProps = ({ vineCompFnCtx, vineFileCtx }, fnItselfNode) => {
  var _a;
  const formalParams = getFunctionParams(fnItselfNode);
  if (formalParams.length === 1) {
    const propsFormalParam = formalParams[0];
    const propsTypeAnnotation = propsFormalParam.typeAnnotation.typeAnnotation;
    vineCompFnCtx.propsAlias = propsFormalParam.name;
    (_a = propsTypeAnnotation.members) == null ? void 0 : _a.forEach((member) => {
      if (!isIdentifier(member.key)) {
        return;
      }
      const propName = member.key.name;
      const propType = vineFileCtx.fileSourceCode.slice(
        member.typeAnnotation.typeAnnotation.start,
        member.typeAnnotation.typeAnnotation.end
      );
      const propMeta = {
        isFromMacroDefine: false,
        isRequired: member.optional === void 0 ? true : !member.optional,
        isBool: propType === "boolean"
      };
      vineCompFnCtx.props[propName] = propMeta;
      vineCompFnCtx.bindings[propName] = VineBindingTypes.PROPS;
    });
  } else if (formalParams.length === 0) {
    const allVinePropMacroCalls = getAllVinePropMacroCall(fnItselfNode);
    allVinePropMacroCalls.forEach(([macroCall, propVarIdentifier]) => {
      var _a2;
      const macroCalleeName = getVineMacroCalleeName(macroCall);
      const propMeta = {
        isFromMacroDefine: true,
        isRequired: macroCalleeName !== "vineProp.optional",
        isBool: false
      };
      const macroCallTypeParamNode = (_a2 = macroCall.typeParameters) == null ? void 0 : _a2.params[0];
      if (macroCallTypeParamNode) {
        const macroCallTypeParam = vineFileCtx.fileSourceCode.slice(
          macroCallTypeParamNode.start,
          macroCallTypeParamNode.end
        );
        propMeta.isBool = macroCallTypeParam === "boolean";
      }
      if (macroCalleeName === "vineProp.withDefault") {
        propMeta.default = macroCall.arguments[0];
        propMeta.validator = macroCall.arguments[1];
      } else {
        propMeta.validator = macroCall.arguments[0];
      }
      const propName = propVarIdentifier.name;
      vineCompFnCtx.props[propName] = propMeta;
      vineCompFnCtx.bindings[propName] = VineBindingTypes.SETUP_REF;
    });
  }
};
const analyzeVineEmits = (analyzeCtx, fnItselfNode) => {
  var _a;
  const { vineCompFnCtx } = analyzeCtx;
  let vineEmitsMacroCall;
  let parentVarDecl;
  traverse(fnItselfNode, {
    enter(descendant, parent) {
      if (isVineMacroOf("vineEmits")(descendant)) {
        vineEmitsMacroCall = descendant;
        const foundVarDeclAncestor = parent.find((ancestor) => isVariableDeclarator(ancestor.node));
        parentVarDecl = foundVarDeclAncestor == null ? void 0 : foundVarDeclAncestor.node;
      }
    }
  });
  if (!vineEmitsMacroCall) {
    return;
  }
  const typeParam = (_a = vineEmitsMacroCall.typeParameters) == null ? void 0 : _a.params[0];
  if (!typeParam) {
    return;
  }
  const emitsTypeLiteralProps = typeParam.members;
  emitsTypeLiteralProps.forEach((prop) => {
    const propName = getTSTypeLiteralPropertySignatureName(prop);
    vineCompFnCtx.emits.push(propName);
  });
  if (parentVarDecl) {
    vineCompFnCtx.emitsAlias = parentVarDecl.id.name;
  }
};
const analyzeVineBindings = (analyzeCtx, fnItselfNode) => {
  const { vineFileCtx, vineCompFnCtx } = analyzeCtx;
  const notContainsMacroStatements = [];
  const fnBody = fnItselfNode.body;
  if (!isBlockStatement(fnBody)) {
    return;
  }
  for (const stmt of fnBody.body) {
    let hasMacroCall = false;
    traverse(stmt, (node) => {
      if (hasMacroCall) {
        return;
      }
      if (isVineMacroCallExpression(node)) {
        hasMacroCall = true;
      }
    });
    if (!hasMacroCall) {
      notContainsMacroStatements.push(stmt);
    }
  }
  for (const stmt of notContainsMacroStatements) {
    const isAllLiteral = analyzeVineFnBodyStmtForBindings(analyzeCtx, stmt);
    if (isAllLiteral) {
      vineCompFnCtx.hoistSetupStmts.push(stmt);
    }
  }
  for (const [importName, {
    isType,
    isNamespace,
    isDefault,
    source
  }] of Object.entries(vineFileCtx.userImports)) {
    if (isType) {
      continue;
    }
    const isSetupConst = isNamespace || isDefault && source.endsWith(".vue") || source === "vue";
    vineCompFnCtx.bindings[importName] = isSetupConst ? VineBindingTypes.SETUP_CONST : VineBindingTypes.SETUP_MAYBE_REF;
  }
  const allTopLevelDeclStmts = vineFileCtx.root.program.body.filter((stmt) => isDeclaration(stmt));
  for (const declStmt of allTopLevelDeclStmts) {
    if (isVineCompFnDecl(declStmt)) {
      const { fnName } = getFunctionInfo(declStmt);
      vineCompFnCtx.bindings[fnName] = VineBindingTypes.SETUP_CONST;
    } else if (isVariableDeclaration(declStmt)) {
      for (const decl of declStmt.declarations) {
        if (isVariableDeclarator(decl) && isIdentifier(decl.id)) {
          vineCompFnCtx.bindings[decl.id.name] = VineBindingTypes.LITERAL_CONST;
        }
      }
    } else if ((isFunctionDeclaration(declStmt) || isClassDeclaration(declStmt)) && declStmt.id) {
      vineCompFnCtx.bindings[declStmt.id.name] = VineBindingTypes.LITERAL_CONST;
    }
  }
};
const analyzeVineStyle = ({ vineFileCtx, vineCompFnCtx }, fnItselfNode) => {
  let vineStyleMacroCall;
  traverse(fnItselfNode, (node) => {
    if (isVineMacroOf("vineStyle")(node)) {
      vineStyleMacroCall = node;
    }
  });
  if (!vineStyleMacroCall) {
    return;
  }
  const macroCalleeName = getVineMacroCalleeName(vineStyleMacroCall);
  const vineStyleArg = vineStyleMacroCall.arguments[0];
  if (!vineStyleArg) {
    return;
  }
  const { styleLang, styleSource, range } = getVineStyleSource(
    vineStyleArg
  );
  const styleMeta = {
    lang: styleLang,
    source: styleSource,
    range,
    scoped: macroCalleeName === "vineStyle.scoped",
    fileCtx: vineFileCtx
  };
  if (vineCompFnCtx.scopeId) {
    vineFileCtx.styleDefine[vineCompFnCtx.scopeId] = styleMeta;
  }
  const cssvarsValueList = parseCssVars([styleSource]);
  if (cssvarsValueList.length > 0) {
    !vineCompFnCtx.cssBindings && (vineCompFnCtx.cssBindings = {});
    cssvarsValueList.forEach((value) => {
      vineCompFnCtx.cssBindings[value] = hashId(`${vineCompFnCtx.fnName}__${value}`);
    });
  }
};
const analyzeVineCustomElement = ({ vineCompFnCtx }, fnItselfNode) => {
  traverse(fnItselfNode, (node) => {
    if (isVineMacroOf("vineCustomElement")(node)) {
      vineCompFnCtx.isCustomElement = true;
    }
  });
};
const analyzeRunners = [
  analyzeVineProps,
  analyzeVineEmits,
  analyzeVineExpose,
  analyzeVineOptions,
  analyzeVineBindings,
  analyzeVineStyle,
  analyzeVineCustomElement
];
function analyzeDifferentKindVineFunctionDecls(analyzeCtx) {
  const { vineCompFnCtx } = analyzeCtx;
  const { fnItselfNode } = vineCompFnCtx;
  if (!fnItselfNode) {
    return;
  }
  analyzeRunners.forEach((exec) => exec(analyzeCtx, fnItselfNode));
}
function analyzeFileImportStmts(vineFileCtx) {
  const { root } = vineFileCtx;
  const fileImportStmts = getImportStatments(root);
  if (!fileImportStmts.length) {
    return;
  }
  for (const importStmt of fileImportStmts) {
    const source = importStmt.source.value;
    const isImportTypeStmt = importStmt.importKind === "type";
    const allSpecifiers = importStmt.specifiers;
    for (const spec of allSpecifiers) {
      const importMeta = {
        source,
        isType: isImportTypeStmt
      };
      if (isImportSpecifier(spec)) {
        const importedName = isStringLiteral(spec.imported) ? spec.imported.value : spec.imported.name;
        const localName = spec.local.name;
        if (spec.importKind === "type") {
          importMeta.isType = true;
        } else if (importedName === "default") {
          importMeta.isDefault = true;
        }
        if (source === "vue") {
          vineFileCtx.vueImportAliases[importedName] = localName || importedName;
        }
        vineFileCtx.userImports[localName] = importMeta;
      } else if (isImportNamespaceSpecifier(spec)) {
        importMeta.isNamespace = true;
        vineFileCtx.userImports[spec.local.name] = importMeta;
      } else if (isImportDefaultSpecifier(spec)) {
        importMeta.isDefault = true;
        vineFileCtx.userImports[spec.local.name] = importMeta;
      }
      const isUsedInTemplate = vineFileCtx.vineCompFns.some(
        (vineCompFn) => isImportUsed(vineCompFn, spec.local.name)
      );
      importMeta.isUsedInTemplate = isUsedInTemplate;
    }
  }
  const lastImportStmt = fileImportStmts[fileImportStmts.length - 1];
  vineFileCtx.importsLastLine = lastImportStmt.loc;
}
function buildVineCompFnCtx(vineCompilerHooks, vineFileCtx, fnDeclNode) {
  var _a, _b;
  const { fnName, fnItselfNode } = getFunctionInfo(fnDeclNode);
  const {
    templateReturn,
    templateStringNode
  } = findVineTagTemplateStringReturn(
    fnDeclNode
  );
  const scopeId = hashId(`${vineFileCtx.fileId}:${fnName}`);
  const templateStringQuasiNode = templateStringNode == null ? void 0 : templateStringNode.quasi.quasis[0];
  const templateSource = (_a = templateStringQuasiNode == null ? void 0 : templateStringQuasiNode.value.raw) != null ? _a : "";
  const vineCompFnCtx = {
    isExport: isExportDeclaration(fnDeclNode),
    isAsync: (_b = fnItselfNode == null ? void 0 : fnItselfNode.async) != null ? _b : false,
    isCustomElement: false,
    fnName,
    scopeId,
    fnDeclNode,
    fnItselfNode,
    templateStringNode,
    templateReturn,
    templateSource,
    propsAlias: "props",
    emitsAlias: "emits",
    props: {},
    emits: [],
    bindings: {},
    cssBindings: {},
    hoistSetupStmts: []
  };
  const analyzeCtx = {
    vineCompilerHooks,
    vineFileCtx,
    vineCompFnCtx
  };
  analyzeDifferentKindVineFunctionDecls(analyzeCtx);
  return vineCompFnCtx;
}
function analyzeVine(vineCompilerHooks, vineFileCtx, vineCompFnDecls) {
  analyzeFileImportStmts(vineFileCtx);
  vineCompFnDecls.forEach(
    (vineFnCompDecl) => {
      vineFileCtx.vineCompFns.push(
        buildVineCompFnCtx(
          vineCompilerHooks,
          vineFileCtx,
          vineFnCompDecl
        )
      );
    }
  );
  const makeErrorOnRefHoistedIdentifiers = (vineFnComp, identifiers) => {
    for (const id of identifiers) {
      const binding = vineFnComp.bindings[id.name];
      if (binding && binding !== VineBindingTypes.LITERAL_CONST) {
        vineCompilerHooks.onError(
          vineWarn(vineFileCtx, {
            msg: `Cannot reference "${id.name}" locally declared variables because it will be hoisted outside of the setup() function.`,
            location: id.loc
          })
        );
      }
    }
  };
  for (const vineFnComp of vineFileCtx.vineCompFns) {
    const allValidatorFnBodys = Object.entries(vineFnComp.props).filter(([_, propMeta]) => Boolean(propMeta.validator)).map(([_, propMeta]) => propMeta.validator.body);
    for (const validatorFnBody of allValidatorFnBodys) {
      const identifiers = [];
      walkIdentifiers(validatorFnBody, (id) => identifiers.push(id));
      makeErrorOnRefHoistedIdentifiers(vineFnComp, identifiers);
    }
    const vineOptionsArg = vineFnComp.options;
    if (vineOptionsArg) {
      const identifiers = [];
      walkIdentifiers(vineOptionsArg, (id) => identifiers.push(id));
      makeErrorOnRefHoistedIdentifiers(vineFnComp, identifiers);
    }
  }
}

function traverseNode(node, filter, callback) {
  const children = node.children;
  if (!filter(node)) {
    return;
  }
  callback(node);
  if (Array.isArray(children)) {
    for (const childNode of children) {
      traverseNode(childNode, filter, callback);
    }
  }
}
function traverseTemplate(templateAstRootNode, filter, callback) {
  for (const childNode of templateAstRootNode.children) {
    traverseNode(childNode, filter, callback);
  }
}

function isTemplateElementNode(node) {
  return node.type === 1;
}
function isComponentNode(node) {
  return node.type === 1 && node.tagType === 1;
}

function createStyleImportStmt(vineFileCtx, vineCompFnCtx, styleDefine) {
  const styleLangExt = getStyleLangFileExt(styleDefine.lang);
  return `import ${showIf(
    Boolean(vineCompFnCtx.isCustomElement),
    `__${vineCompFnCtx.fnName.toLowerCase()}_styles from `
  )}'${vineFileCtx.fileId.replace(/\.vine\.ts$/, "")}?type=vine-style&scopeId=${vineCompFnCtx.scopeId}&comp=${vineCompFnCtx.fnName}&lang=${styleDefine.lang}${showIf(
    Boolean(styleDefine.scoped),
    "&scoped=true"
  )}&virtual.${styleLangExt}';`;
}

function topoSort(relationsMap) {
  const visited = {};
  const sorted = [];
  function dfs(node, stack) {
    if (stack[node]) {
      return null;
    }
    if (visited[node])
      return;
    visited[node] = true;
    stack[node] = true;
    for (const depNode of relationsMap[node]) {
      const result = dfs(depNode, stack);
      if (result === null) {
        return null;
      }
    }
    stack[node] = false;
    sorted.push(node);
  }
  for (const node of Object.keys(relationsMap)) {
    dfs(node, {});
  }
  if (sorted.length !== Object.keys(visited).length) {
    return null;
  }
  return sorted;
}
function sortStyleImport(vineFileCtx) {
  var _a, _b;
  const { vineCompFns, styleDefine } = vineFileCtx;
  const relationsMap = Object.fromEntries(
    vineCompFns.map(
      (compFnCtx) => [compFnCtx.fnName, /* @__PURE__ */ new Set()]
    )
  );
  for (const compFnCtx of vineCompFns) {
    const templateAstRootNode = compFnCtx.templateAst;
    if (!templateAstRootNode) {
      continue;
    }
    traverseTemplate(
      templateAstRootNode,
      (templateAstNode) => isTemplateElementNode(templateAstNode) && isComponentNode(templateAstNode),
      (templateAstNode) => {
        relationsMap[compFnCtx.fnName].add(
          templateAstNode.tag
        );
      }
    );
  }
  const sortedStyleImportStmts = ((_b = (_a = topoSort(relationsMap)) == null ? void 0 : _a.map(
    (compName) => vineCompFns.find(
      (compFn) => compFn.fnName === compName
    )
  )) != null ? _b : []).filter(
    (fnCompCtx) => Boolean(styleDefine[fnCompCtx.scopeId])
  ).map(
    (fnCompCtx) => [fnCompCtx, styleDefine[fnCompCtx.scopeId]]
  ).map(
    ([fnCompCtx, styleMeta]) => createStyleImportStmt(
      vineFileCtx,
      fnCompCtx,
      styleMeta
    )
  );
  return sortedStyleImportStmts;
}

var toString = {}.toString;

var isarray = Array.isArray || function (arr) {
  return toString.call(arr) == '[object Array]';
};

/*!
 * isobject <https://github.com/jonschlinkert/isobject>
 *
 * Copyright (c) 2014-2015, Jon Schlinkert.
 * Licensed under the MIT License.
 */

var isArray$1 = isarray;

var isobject = function isObject(val) {
  return val != null && typeof val === 'object' && isArray$1(val) === false;
};

/**
 * line-column - Convert efficiently index to/from line-column in a string
 * @module  lineColumn
 * @license MIT
 */

var isArray  = isarray;
var isObject = isobject;

var lineColumn = LineColumnFinder;

/**
 * Finder for index and line-column from given string.
 *
 * You can call this without `new` operator as it returns an instance anyway.
 *
 * @class
 * @param {string} str - A string to be parsed.
 * @param {Object|number} [options] - Options.
 *     This can be an index in the string for shorthand of `lineColumn(str, index)`.
 * @param {number} [options.origin=1] - The origin value of line and column.
 */
function LineColumnFinder(str, options) {
  if (!(this instanceof LineColumnFinder)) {
    if (typeof options === "number") {
      return (new LineColumnFinder(str)).fromIndex(options);
    }
    return new LineColumnFinder(str, options);
  }

  this.str = str || "";
  this.lineToIndex = buildLineToIndex(this.str);

  options = options || {};
  this.origin = typeof options.origin === "undefined" ? 1 : options.origin;
}

/**
 * Find line and column from index in the string.
 *
 * @param  {number} index - Index in the string. (0-origin)
 * @return {Object|null}
 *     Found line number and column number in object `{ line: X, col: Y }`.
 *     If the given index is out of range, it returns `null`.
 */
LineColumnFinder.prototype.fromIndex = function (index) {
  if (index < 0 || index >= this.str.length || isNaN(index)) {
    return null;
  }

  var line = findLowerIndexInRangeArray(index, this.lineToIndex);
  return {
    line: line + this.origin,
    col:  index - this.lineToIndex[line] + this.origin
  };
};

/**
 * Find index from line and column in the string.
 *
 * @param  {number|Object|Array} line - Line number in the string.
 *     This can be an Object of `{ line: X, col: Y }`, or
 *     an Array of `[line, col]`.
 * @param  {number} [column] - Column number in the string.
 *     This must be omitted or undefined when Object or Array is given
 *     to the first argument.
 * @return {number}
 *     Found index in the string. (always 0-origin)
 *     If the given line or column is out of range, it returns `-1`.
 */
LineColumnFinder.prototype.toIndex = function (line, column) {
  if (typeof column === "undefined") {
    if (isArray(line) && line.length >= 2) {
      return this.toIndex(line[0], line[1]);
    }
    if (isObject(line) && "line" in line && ("col" in line || "column" in line)) {
      return this.toIndex(line.line, ("col" in line ? line.col : line.column));
    }
    return -1;
  }
  if (isNaN(line) || isNaN(column)) {
    return -1;
  }

  line -= this.origin;
  column -= this.origin;

  if (line >= 0 && column >= 0 && line < this.lineToIndex.length) {
    var lineIndex = this.lineToIndex[line];
    var nextIndex = (
      line === this.lineToIndex.length - 1
        ? this.str.length
        : this.lineToIndex[line + 1]
    );

    if (column < nextIndex - lineIndex) {
      return lineIndex + column;
    }
  }
  return -1;
};

/**
 * Build an array of indexes of each line from a string.
 *
 * @private
 * @param   str {string}  An input string.
 * @return  {number[]}    Built array of indexes. The key is line number.
 */
function buildLineToIndex(str) {
  var lines = str.split("\n"),
      lineToIndex = new Array(lines.length),
      index = 0;

  for (var i = 0, l = lines.length; i < l; i++) {
    lineToIndex[i] = index;
    index += lines[i].length + /* "\n".length */ 1;
  }
  return lineToIndex;
}

/**
 * Find a lower-bound index of a value in a sorted array of ranges.
 *
 * Assume `arr = [0, 5, 10, 15, 20]` and
 * this returns `1` for `value = 7` (5 <= value < 10),
 * and returns `3` for `value = 18` (15 <= value < 20).
 *
 * @private
 * @param  arr   {number[]} An array of values representing ranges.
 * @param  value {number}   A value to be searched.
 * @return {number} Found index. If not found `-1`.
 */
function findLowerIndexInRangeArray(value, arr) {
  if (value >= arr[arr.length - 1]) {
    return arr.length - 1;
  }

  var min = 0, max = arr.length - 2, mid;
  while (min < max) {
    mid = min + ((max - min) >> 1);

    if (value < arr[mid]) {
      max = mid - 1;
    } else if (value >= arr[mid + 1]) {
      min = mid + 1;
    } else { // value >= arr[mid] && value < arr[mid + 1]
      min = mid;
      break;
    }
  }
  return min;
}

var lineColumn$1 = /*@__PURE__*/getDefaultExportFromCjs(lineColumn);

var __defProp$4 = Object.defineProperty;
var __getOwnPropSymbols$4 = Object.getOwnPropertySymbols;
var __hasOwnProp$4 = Object.prototype.hasOwnProperty;
var __propIsEnum$4 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$4 = (obj, key, value) => key in obj ? __defProp$4(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$4 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$4.call(b, prop))
      __defNormalProp$4(a, prop, b[prop]);
  if (__getOwnPropSymbols$4)
    for (var prop of __getOwnPropSymbols$4(b)) {
      if (__propIsEnum$4.call(b, prop))
        __defNormalProp$4(a, prop, b[prop]);
    }
  return a;
};
function babelParse(code, options = {}) {
  const finalOptions = __spreadValues$4({
    sourceType: "module",
    plugins: ["typescript"],
    errorRecovery: true
  }, options);
  try {
    return parse(code, finalOptions);
  } catch (err) {
    return parse("", finalOptions);
  }
}

var __defProp$3 = Object.defineProperty;
var __getOwnPropSymbols$3 = Object.getOwnPropertySymbols;
var __hasOwnProp$3 = Object.prototype.hasOwnProperty;
var __propIsEnum$3 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$3 = (obj, key, value) => key in obj ? __defProp$3(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$3 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$3.call(b, prop))
      __defNormalProp$3(a, prop, b[prop]);
  if (__getOwnPropSymbols$3)
    for (var prop of __getOwnPropSymbols$3(b)) {
      if (__propIsEnum$3.call(b, prop))
        __defNormalProp$3(a, prop, b[prop]);
    }
  return a;
};
function compileVineTemplate(source, params) {
  return compile(source, __spreadValues$3({
    mode: "module",
    hoistStatic: true,
    cacheHandlers: true,
    inline: true
  }, params));
}
function saveImportSpecifier(mergedImportsMap, importStmts) {
  for (const importStmt of importStmts) {
    const allImportSpecs = importStmt.specifiers;
    const source = importStmt.source.value;
    for (const spec of allImportSpecs) {
      if (isImportSpecifier(spec)) {
        const specName = isIdentifier(spec.imported) ? spec.imported.name : spec.imported.value;
        const specAlias = isIdentifier(spec.local) ? spec.local.name : specName;
        const specMeta = mergedImportsMap.get(source);
        if (!specMeta) {
          mergedImportsMap.set(
            source,
            {
              type: "namedSpecifier",
              specs: /* @__PURE__ */ new Map([[specName, specAlias]])
            }
          );
        } else {
          specMeta.specs.set(specName, specAlias);
        }
      } else if (isImportDefaultSpecifier(spec)) {
        mergedImportsMap.set(
          source,
          {
            type: "defaultSpecifier",
            localName: spec.local.name
          }
        );
      } else {
        mergedImportsMap.set(
          source,
          {
            type: "namespaceSpecifier",
            localName: spec.local.name
          }
        );
      }
    }
  }
}
function isExportRenderFnNode(node) {
  var _a;
  if (!isExportNamedDeclaration(node)) {
    return false;
  }
  return isFunctionDeclaration(node.declaration) && ((_a = node.declaration.id) == null ? void 0 : _a.name) === "render";
}
function getQuasisNode(vineFnCompCtx) {
  var _a;
  return (_a = vineFnCompCtx.templateStringNode) == null ? void 0 : _a.quasi.quasis[0];
}
function computeTemplateErrLocation(vineFileCtx, vineFnCompCtx, errLocation) {
  const quasisNode = getQuasisNode(vineFnCompCtx);
  if (!quasisNode) {
    return;
  }
  const mapper = lineColumn$1(
    vineFileCtx.fileSourceCode.original,
    { origin: 1 }
  );
  const templateStart = quasisNode.start;
  const startOffset = templateStart + errLocation.start.offset;
  const endOffset = templateStart + errLocation.end.offset;
  const start = mapper.fromIndex(startOffset);
  const end = mapper.fromIndex(endOffset);
  if (!start || !end) {
    return;
  }
  const [startColumn, endColumn] = start.line === end.line && start.col === end.col ? [1, Number.POSITIVE_INFINITY] : [start.col, end.col];
  const loc = {
    identifierName: errLocation.source,
    filename: vineFileCtx.fileId,
    start: {
      line: start.line,
      column: startColumn,
      index: startOffset
    },
    end: {
      line: end.line,
      column: endColumn,
      index: endOffset
    }
  };
  return loc;
}
function createSeparatedTemplateComposer(compilerHooks) {
  const templateCompileResults = /* @__PURE__ */ new WeakMap();
  const generatedPreambleStmts = /* @__PURE__ */ new WeakMap();
  return {
    templateCompileResults,
    generatedPreambleStmts,
    compileSetupFnReturns: ({
      vineFileCtx,
      vineCompFnCtx: vineFnCompCtx,
      templateSource,
      mergedImportsMap,
      bindingMetadata
    }) => {
      let hasTemplateCompileErr = false;
      const compileResult = compileVineTemplate(
        templateSource,
        {
          scopeId: `data-v-${vineFnCompCtx.scopeId}`,
          bindingMetadata,
          inline: false,
          onError: (e) => {
            if (hasTemplateCompileErr) {
              return;
            }
            hasTemplateCompileErr = true;
            compilerHooks.onError(
              vineErr(
                vineFileCtx,
                {
                  msg: `[Vine template compile error] ${e.message}`,
                  location: e.loc && computeTemplateErrLocation(vineFileCtx, vineFnCompCtx, e.loc)
                }
              )
            );
          },
          onWarn: (e) => {
            compilerHooks.onWarn(
              vineWarn(
                vineFileCtx,
                {
                  msg: `[Vine template compile warning] ${e.message}`,
                  location: e.loc && computeTemplateErrLocation(vineFileCtx, vineFnCompCtx, e.loc)
                }
              )
            );
          }
        }
      );
      vineFnCompCtx.templateAst = hasTemplateCompileErr ? void 0 : parse$1(templateSource);
      if (hasTemplateCompileErr) {
        return "";
      }
      const { code } = compileResult;
      const generatedCodeAst = babelParse(code);
      saveImportSpecifier(
        mergedImportsMap,
        generatedCodeAst.program.body.filter(
          (stmt) => isImportDeclaration(stmt)
        )
      );
      let exportRenderFnNode;
      for (const codeStmt of generatedCodeAst.program.body) {
        if (isImportDeclaration(codeStmt)) {
          continue;
        } else if (isExportRenderFnNode(codeStmt)) {
          exportRenderFnNode = codeStmt;
          continue;
        }
        appendToMapArray(
          generatedPreambleStmts,
          vineFnCompCtx,
          code.slice(
            codeStmt.start,
            codeStmt.end
          )
        );
      }
      if (!exportRenderFnNode) {
        compilerHooks.onError(vineErr(vineFileCtx, {
          msg: "[Vine Error] Cannot find export render function on template composing"
        }));
        return "";
      }
      templateCompileResults.set(
        vineFnCompCtx,
        code.slice(
          exportRenderFnNode.start,
          exportRenderFnNode.end
        ).replace(
          "export function render",
          "function __sfc_render"
        )
      );
      const allBindings = __spreadValues$3({}, bindingMetadata);
      for (const key in vineFileCtx.userImports) {
        if (!vineFileCtx.userImports[key].isType && vineFileCtx.userImports[key].isUsedInTemplate) {
          allBindings[key] = true;
        }
      }
      let setupFnReturns = "{ ";
      for (const key in allBindings) {
        if (allBindings[key] === true && vineFileCtx.userImports[key].source !== "vue" && !vineFileCtx.userImports[key].source.endsWith(".vue")) {
          setupFnReturns += `get ${key}() { return ${key} }, `;
        } else if (bindingMetadata[key] === VineBindingTypes.SETUP_LET) {
          const setArg = key === "v" ? "_v" : "v";
          setupFnReturns += `get ${key}() { return ${key} }, set ${key}(${setArg}) { ${key} = ${setArg} }, `;
        } else if (bindingMetadata[key] === VineBindingTypes.PROPS) ; else {
          setupFnReturns += `${key}, `;
        }
      }
      setupFnReturns = `${setupFnReturns.replace(/, $/, "")} }`;
      return setupFnReturns;
    }
  };
}
function createInlineTemplateComposer(compilerHooks) {
  const templateCompileResults = /* @__PURE__ */ new WeakMap();
  const generatedPreambleStmts = /* @__PURE__ */ new WeakMap();
  return {
    templateCompileResults,
    generatedPreambleStmts,
    compileSetupFnReturns: ({
      vineFileCtx,
      vineCompFnCtx: vineFnCompCtx,
      templateSource,
      mergedImportsMap,
      bindingMetadata
    }) => {
      let hasTemplateCompileErr = false;
      const compileResult = compileVineTemplate(
        templateSource,
        {
          scopeId: `data-v-${vineFnCompCtx.scopeId}`,
          bindingMetadata,
          onError: (e) => {
            if (hasTemplateCompileErr) {
              return;
            }
            hasTemplateCompileErr = true;
            compilerHooks.onError(
              vineErr(
                vineFileCtx,
                {
                  msg: `[Vine template compile error] ${e.message}`,
                  location: e.loc && computeTemplateErrLocation(vineFileCtx, vineFnCompCtx, e.loc)
                }
              )
            );
          },
          onWarn: (e) => {
            compilerHooks.onWarn(
              vineWarn(
                vineFileCtx,
                {
                  msg: `[Vine template compile warning] ${e.message}`,
                  location: e.loc && computeTemplateErrLocation(vineFileCtx, vineFnCompCtx, e.loc)
                }
              )
            );
          }
        }
      );
      vineFnCompCtx.templateAst = hasTemplateCompileErr ? void 0 : parse$1(templateSource);
      if (hasTemplateCompileErr) {
        return "";
      }
      const { preamble, code } = compileResult;
      const preambleAst = babelParse(preamble);
      saveImportSpecifier(
        mergedImportsMap,
        preambleAst.program.body.filter(
          (stmt) => isImportDeclaration(stmt)
        )
      );
      const preambleStmts = preambleAst.program.body.filter(
        (stmt) => !isImportDeclaration(stmt)
      );
      preambleStmts.forEach((stmt) => appendToMapArray(
        generatedPreambleStmts,
        vineFnCompCtx,
        preamble.slice(
          stmt.start,
          stmt.end
        )
      ));
      templateCompileResults.set(vineFnCompCtx, code);
      return code;
    }
  };
}

function findIdentifierFromExp(scriptContent) {
  const identifiers = [];
  const ast = babelParse(scriptContent);
  traverse(ast, (node) => {
    if (node.type === "Identifier") {
      identifiers.push(node);
    }
  });
  return identifiers;
}
function genCSSVarsList(cssBindings, propsAlias, bindings, inline = false) {
  let res = "";
  if (cssBindings) {
    for (const cssBindScriptContent in cssBindings) {
      const cssBindValue = cssBindings[cssBindScriptContent];
      const ms = new MagicString(cssBindScriptContent);
      const cssBindKeySgNodes = findIdentifierFromExp(cssBindScriptContent);
      cssBindKeySgNodes.forEach((node) => {
        ms.overwrite(
          node.start,
          node.end,
          inline ? genCSSVarsValue(ms, node, bindings, propsAlias) : `_ctx.${cssBindScriptContent.slice(
            node.start,
            node.end
          )}`
        );
      });
      res = `${res}  '${cssBindValue}': (${ms.toString()}),
`;
    }
  }
  return res;
}
function genCSSVarsValue(ms, node, bindings, propsAlias) {
  let res = "";
  const nodeContent = ms.original.slice(
    node.start,
    node.end
  );
  for (const bindingsKey in bindings) {
    const bindingValue = bindings[bindingsKey];
    if (nodeContent === bindingsKey) {
      switch (bindingValue) {
        case VineBindingTypes.PROPS:
        case VineBindingTypes.PROPS_ALIASED:
          res = `${propsAlias}.${nodeContent}`;
          break;
        case VineBindingTypes.SETUP_CONST:
        case VineBindingTypes.SETUP_REACTIVE_CONST:
        case VineBindingTypes.LITERAL_CONST:
          res = nodeContent;
          break;
        case VineBindingTypes.SETUP_MAYBE_REF:
        case VineBindingTypes.SETUP_LET:
          res = `_unref(${nodeContent})`;
          break;
        case VineBindingTypes.SETUP_REF:
          res = `${nodeContent}.value`;
          break;
        default:
          res = `_ctx.${nodeContent}`;
      }
    }
  }
  return res;
}
function compileCSSVars(vineCompFnCtx, inline = false) {
  const {
    cssBindings,
    propsAlias,
    bindings
  } = vineCompFnCtx;
  if (!cssBindings)
    return "";
  const varList = genCSSVarsList(
    cssBindings,
    propsAlias,
    bindings,
    inline
  );
  return `_${CSS_VARS_HELPER}(_ctx => ({
${varList}
}))`;
}

function wrapWithAsyncContext(isNeedResult, exprSourceCode) {
  return isNeedResult ? `;(
    ([__temp,__restore] = _withAsyncContext(() => ${exprSourceCode})),
    await __temp,
    __restore()
  );` : `;(
    ([__temp,__restore] = _withAsyncContext(() => ${exprSourceCode})),
    __temp = await __temp,
    __restore(),
    __temp
  );`;
}
function mayContainAwaitExpr(targetNode) {
  let awaitExpr;
  let isMayCotainAwaitExpr = isVariableDeclaration(targetNode) || isAssignmentExpression(targetNode) || isExpressionStatement(targetNode);
  if (!isMayCotainAwaitExpr) {
    return false;
  }
  traverse(targetNode, (descendant) => {
    if (isAwaitExpression(descendant)) {
      isMayCotainAwaitExpr = true;
      awaitExpr = descendant;
    }
  });
  return isMayCotainAwaitExpr && awaitExpr;
}
function registerImport(mergedImportsMap, importSource, specifier, alias) {
  const vueVineImports = mergedImportsMap.get(importSource);
  if (!vueVineImports) {
    mergedImportsMap.set(importSource, {
      type: "namedSpecifier",
      specs: /* @__PURE__ */ new Map([[specifier, alias]])
    });
  } else {
    const vueImportsSpecs = vueVineImports.specs;
    vueImportsSpecs.set(specifier, alias);
  }
}
function transformFile(vineFileCtx, compilerHooks, inline = true) {
  var _a, _b, _c;
  const isDev = compilerHooks.getCompilerCtx().options.mode !== "production";
  const ms = vineFileCtx.fileSourceCode;
  const styleImportStmts = sortStyleImport(vineFileCtx);
  const mergedImportsMap = /* @__PURE__ */ new Map();
  const {
    templateCompileResults,
    generatedPreambleStmts,
    compileSetupFnReturns
  } = inline ? createInlineTemplateComposer(compilerHooks) : createSeparatedTemplateComposer(compilerHooks);
  let isPrependedUseDefaults = false;
  for (const vineCompFnCtx of vineFileCtx.vineCompFns) {
    const setupFnReturns = compileSetupFnReturns({
      vineFileCtx,
      vineCompFnCtx,
      templateSource: vineCompFnCtx.templateSource,
      mergedImportsMap,
      bindingMetadata: vineCompFnCtx.bindings
    });
    let vueImportsMeta = mergedImportsMap.get("vue");
    if (!vueImportsMeta) {
      vueImportsMeta = {
        type: "namedSpecifier",
        specs: /* @__PURE__ */ new Map()
      };
      mergedImportsMap.set("vue", vueImportsMeta);
    }
    const vueImportsSpecs = vueImportsMeta.specs;
    if (!vueImportsSpecs.has(DEFINE_COMPONENT_HELPER)) {
      vueImportsSpecs.set(DEFINE_COMPONENT_HELPER, `_${DEFINE_COMPONENT_HELPER}`);
    }
    if (!vueImportsSpecs.has(CSS_VARS_HELPER) && vineCompFnCtx.cssBindings) {
      vueImportsSpecs.set(CSS_VARS_HELPER, `_${CSS_VARS_HELPER}`);
      inline && vueImportsSpecs.set(UN_REF_HELPER, `_${UN_REF_HELPER}`);
    }
    const vineCompFnStart = vineCompFnCtx.fnDeclNode.start;
    const vineCompFnEnd = vineCompFnCtx.fnDeclNode.end;
    const vineCompFnBody = vineCompFnCtx.fnItselfNode.body;
    if (isBlockStatement(vineCompFnBody)) {
      let hasAwait = false;
      for (const vineFnBodyStmt of vineCompFnBody.body) {
        const isContained = mayContainAwaitExpr(vineFnBodyStmt);
        if (!isContained) {
          continue;
        }
        hasAwait = true;
        ms.update(
          vineFnBodyStmt.start,
          vineFnBodyStmt.end,
          wrapWithAsyncContext(
            isExpressionStatement(vineFnBodyStmt),
            ms.original.slice(
              vineFnBodyStmt.start,
              vineFnBodyStmt.end
            )
          )
        );
      }
      const firstStmt = vineCompFnBody.body[0];
      const lastStmt = vineCompFnBody.body[vineCompFnBody.body.length - 1];
      ms.remove(vineCompFnStart, firstStmt.start);
      ms.remove(lastStmt.end, vineCompFnEnd);
      vineCompFnBody.body.forEach((stmt) => {
        if (isStatementContainsVineMacroCall(stmt) || isReturnStatement(stmt)) {
          ms.remove(stmt.start, stmt.end);
        }
      });
      const setupCtxDestructFormalParams = [];
      if (vineCompFnCtx.emits.length > 0) {
        setupCtxDestructFormalParams.push({
          field: "emits",
          alias: vineCompFnCtx.emitsAlias
        });
      }
      if (vineCompFnCtx.expose) {
        setupCtxDestructFormalParams.push({
          field: "expose"
        });
      }
      let setupFormalParams = `__props${setupCtxDestructFormalParams.length > 0 ? `, { ${setupCtxDestructFormalParams.map(
        ({ field, alias }) => `${field}${showIf(Boolean(alias), `: ${alias}`)}`
      ).join(", ")} }` : " /* No setup ctx destructuring */"}`;
      if (Object.values(vineCompFnCtx.props).some((meta) => Boolean(meta.isFromMacroDefine))) {
        registerImport(
          mergedImportsMap,
          "vue",
          TO_REFS_HELPER,
          `_${TO_REFS_HELPER}`
        );
        const propsFromMacro = Object.entries(vineCompFnCtx.props).filter(([_, meta]) => Boolean(meta.isFromMacroDefine)).map(([propName, _]) => propName);
        ms.prependLeft(
          firstStmt.start,
          `const { ${propsFromMacro.join(", ")} } = _toRefs(__props);
`
        );
      }
      if (Array.from(
        Object.entries(vineCompFnCtx.cssBindings)
      ).length > 0) {
        ms.prependLeft(
          firstStmt.start,
          `
${compileCSSVars(vineCompFnCtx, inline)}
`
        );
      }
      let propsDeclarationStmt = `const ${vineCompFnCtx.propsAlias} = __props;`;
      if (Object.values(vineCompFnCtx.props).some((meta) => Boolean(meta.default)) && !isPrependedUseDefaults) {
        isPrependedUseDefaults = true;
        registerImport(
          mergedImportsMap,
          "vue-vine",
          USE_DEFAULTS_HELPER,
          `_${USE_DEFAULTS_HELPER}`
        );
        propsDeclarationStmt = `const ${vineCompFnCtx.propsAlias} = _useDefaults(__props, {
${Object.entries(vineCompFnCtx.props).filter(([_, propMeta]) => Boolean(propMeta.default)).map(([propName, propMeta]) => `  ${propName}: ${ms.original.slice(
          propMeta.default.start,
          propMeta.default.end
        )}`).join(",\n")}
})
`;
      }
      ms.prependLeft(
        firstStmt.start,
        propsDeclarationStmt
      );
      if (hasAwait) {
        ms.prependLeft(
          firstStmt.start,
          "let __temp, __restore;\n"
        );
      }
      ms.appendRight(lastStmt.end, `
return ${setupFnReturns};`);
      ms.prependLeft(firstStmt.start, `setup(${setupFormalParams}) {
`);
      ms.appendRight(lastStmt.end, "\n}");
      ms.prependLeft(firstStmt.start, `const __vine = _defineComponent({
${vineCompFnCtx.options ? `...${ms.original.slice(
        vineCompFnCtx.options.start,
        vineCompFnCtx.options.end
      )},` : `name: '${vineCompFnCtx.fnName}',`}
${Object.keys(vineCompFnCtx.props).length > 0 ? `props: {
${Object.entries(vineCompFnCtx.props).map(([propName, propMeta]) => {
        const metaFields = [];
        if (propMeta.isRequired) {
          metaFields.push("required: true");
        }
        if (propMeta.isBool) {
          metaFields.push("type: Boolean");
        }
        if (propMeta.validator) {
          metaFields.push(`validator: ${ms.original.slice(
            propMeta.validator.start,
            propMeta.validator.end
          )}`);
        }
        return `${propName}: { ${showIf(
          metaFields.filter(Boolean).length > 0,
          filterJoin(metaFields, ", "),
          "/* Simple prop */"
        )} },`;
      }).join("\n")}
},` : "/* No props */"}
${vineCompFnCtx.emits.length > 0 ? `emits: [${vineCompFnCtx.emits.map((emit) => `'${emit}'`).join(", ")}],` : "/* No emits */"}
`);
      ms.appendRight(lastStmt.end, "\n})");
      ms.prependLeft(firstStmt.start, `
export const ${vineCompFnCtx.fnName} = (() => {
${(_b = (_a = generatedPreambleStmts.get(vineCompFnCtx)) == null ? void 0 : _a.join("\n")) != null ? _b : ""}
`);
      ms.appendRight(
        lastStmt.end,
        `
${inline ? "" : `${(_c = templateCompileResults.get(vineCompFnCtx)) != null ? _c : ""}
__vine.render = __sfc_render`}
${showIf(
          Boolean(vineFileCtx.styleDefine[vineCompFnCtx.scopeId]),
          `__vine.__scopeId = 'data-v-${vineCompFnCtx.scopeId}';
${isDev ? `__vine.__hmrId = '${vineCompFnCtx.scopeId}';` : ""}
`
        )}${showIf(
          Boolean(vineCompFnCtx.isCustomElement),
          `__vine.styles = [__${vineCompFnCtx.fnName.toLowerCase()}_styles];
`
        )}
return __vine
})();`
      );
      if (isDev) {
        ms.appendRight(
          ms.length(),
          `
typeof __VUE_HMR_RUNTIME__ !== "undefined" && __VUE_HMR_RUNTIME__.createRecord(${vineCompFnCtx.fnName}.__hmrId, ${vineCompFnCtx.fnName});
`
        );
      }
    }
  }
  if (isDev) {
    ms.appendRight(
      ms.length(),
      `export const _rerender_only = ${vineFileCtx.renderOnly}
export const _rerender_vcf_fn_name = ${vineFileCtx.hmrCompFnsName ? `"${vineFileCtx.hmrCompFnsName}"` : '""'}
import.meta.hot.accept((mod) => {
if (!mod){return;}
const { _rerender_only, _rerender_vcf_fn_name } = mod;
if (!_rerender_vcf_fn_name){return;}
const component = mod[_rerender_vcf_fn_name];
if (_rerender_only) {
  __VUE_HMR_RUNTIME__.rerender(component.__hmrId, component.render);
} else {
  __VUE_HMR_RUNTIME__.reload(component.__hmrId, component);
}
});`
    );
  }
  ms.prepend(`
${styleImportStmts.join("\n")}
`);
  for (const [importSource, importMeta] of mergedImportsMap) {
    if (importMeta.type === "namedSpecifier") {
      const { specs } = importMeta;
      const specifiers = [...specs.entries()];
      const importStmt = `import { ${specifiers.map(
        ([specifier, alias]) => `${specifier}${showIf(Boolean(alias), ` as ${alias}`)}`
      ).join(", ")} } from '${importSource}';
`;
      ms.prepend(importStmt);
    } else if (importMeta.type === "defaultSpecifier") {
      const importStmt = `import ${importMeta.localName} from '${importSource}';
`;
      ms.prepend(importStmt);
    }
  }
}

var __defProp$2 = Object.defineProperty;
var __defProps$1 = Object.defineProperties;
var __getOwnPropDescs$1 = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols$2 = Object.getOwnPropertySymbols;
var __hasOwnProp$2 = Object.prototype.hasOwnProperty;
var __propIsEnum$2 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$2 = (obj, key, value) => key in obj ? __defProp$2(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$2 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$2.call(b, prop))
      __defNormalProp$2(a, prop, b[prop]);
  if (__getOwnPropSymbols$2)
    for (var prop of __getOwnPropSymbols$2(b)) {
      if (__propIsEnum$2.call(b, prop))
        __defNormalProp$2(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps$1 = (a, b) => __defProps$1(a, __getOwnPropDescs$1(b));
const scss = (source, map, options, load = require) => {
  const nodeSass = load("sass");
  const finalOptions = __spreadProps$1(__spreadValues$2({}, options), {
    data: source,
    sourceMap: Boolean(map)
  });
  try {
    const result = nodeSass.renderSync(finalOptions);
    const dependencies = result.stats.includedFiles;
    if (map) {
      return {
        code: result.css.toString(),
        map: merge(map, JSON.parse(result.map.toString())),
        errors: [],
        dependencies
      };
    }
    return { code: result.css.toString(), errors: [], dependencies };
  } catch (e) {
    return { code: "", errors: [e], dependencies: [] };
  }
};
const sass = (source, map, options, load = require) => scss(
  source,
  map,
  __spreadProps$1(__spreadValues$2({}, options), {
    indentedSyntax: true
  }),
  load
);
const less = (source, map, options, load = require) => {
  const nodeLess = load("less");
  let result;
  let error = null;
  nodeLess.render(
    source,
    __spreadProps$1(__spreadValues$2({}, options), { syncImport: true }),
    (err, output) => {
      error = err;
      result = output;
    }
  );
  if (error)
    return { code: "", errors: [error], dependencies: [] };
  const dependencies = result.imports;
  if (map) {
    return {
      code: result.css.toString(),
      map: merge(map, result.map),
      errors: [],
      dependencies
    };
  }
  return {
    code: result.css.toString(),
    errors: [],
    dependencies
  };
};
const stylus = (source, map, options, load = require) => {
  const nodeStylus = load("stylus");
  try {
    const ref = nodeStylus(source);
    Object.keys(options).forEach((key) => ref.set(key, options[key]));
    if (map)
      ref.set("sourcemap", { inline: false, comment: false });
    const result = ref.render();
    const dependencies = ref.deps();
    if (map) {
      return {
        code: result,
        map: merge(map, ref.sourcemap),
        errors: [],
        dependencies
      };
    }
    return { code: result, errors: [], dependencies };
  } catch (e) {
    return { code: "", errors: [e], dependencies: [] };
  }
};
const processors = {
  less,
  sass,
  scss,
  stylus
};

const animationNameRE = /^(-\w+-)?animation-name$/;
const animationRE = /^(-\w+-)?animation$/;
const processedRules = /* @__PURE__ */ new WeakSet();
const scopedPlugin = (options) => {
  const { id = "", fileCtx, compilerCtx } = options != null ? options : {};
  const keyframes = /* @__PURE__ */ Object.create(null);
  const shortId = id.replace(/^data-v-/, "");
  function onPluginWarn(params) {
    compilerCtx == null ? void 0 : compilerCtx.vineCompileWarnings.push(
      vineWarn(fileCtx, params)
    );
  }
  function processRule(id2, rule) {
    if (processedRules.has(rule) || rule.parent && rule.parent.type === "atrule" && /-?keyframes$/.test(rule.parent.name)) {
      return;
    }
    processedRules.add(rule);
    rule.selector = selectorParser((selectorRoot) => {
      selectorRoot.each((selector) => {
        rewriteSelector(id2, selector, selectorRoot);
      });
    }).processSync(rule.selector);
  }
  function rewriteSelector(id2, selector, selectorRoot, slotted = false) {
    let node = null;
    let shouldInject = true;
    selector.each((n) => {
      if (n.type === "combinator" && (n.value === ">>>" || n.value === "/deep/")) {
        n.value = " ";
        n.spaces.before = n.spaces.after = "";
        onPluginWarn({
          msg: "the >>> and /deep/ combinators have been deprecated. Use :deep() instead.",
          location: null
        });
        return false;
      }
      if (n.type === "pseudo") {
        const { value } = n;
        if (value === ":deep" || value === "::v-deep") {
          if (n.nodes.length) {
            let last = n;
            n.nodes[0].each((ss) => {
              selector.insertAfter(last, ss);
              last = ss;
            });
            const prev = selector.at(selector.index(n) - 1);
            if (!prev || !isSpaceCombinator(prev)) {
              selector.insertAfter(
                n,
                selectorParser.combinator({
                  value: " "
                })
              );
            }
            selector.removeChild(n);
          } else {
            onPluginWarn({
              msg: "::v-deep usage as a combinator has been deprecated. Use :deep(<inner-selector>) instead.",
              location: null
            });
            const prev = selector.at(selector.index(n) - 1);
            if (prev && isSpaceCombinator(prev)) {
              selector.removeChild(prev);
            }
            selector.removeChild(n);
          }
          return false;
        }
        if (value === ":slotted" || value === "::v-slotted") {
          rewriteSelector(id2, n.nodes[0], selectorRoot, true);
          let last = n;
          n.nodes[0].each((ss) => {
            selector.insertAfter(last, ss);
            last = ss;
          });
          selector.removeChild(n);
          shouldInject = false;
          return false;
        }
        if (value === ":global" || value === "::v-global") {
          selectorRoot.insertAfter(selector, n.nodes[0]);
          selectorRoot.removeChild(selector);
          return false;
        }
      }
      if (n.type !== "pseudo" && n.type !== "combinator") {
        node = n;
      }
    });
    if (node) {
      node.spaces.after = "";
    } else {
      selector.first.spaces.before = "";
    }
    if (shouldInject) {
      const idToAdd = slotted ? `${id2}-s` : id2;
      selector.insertAfter(
        node,
        selectorParser.attribute({
          attribute: idToAdd,
          value: idToAdd,
          raws: {},
          quoteMark: '"'
        })
      );
    }
  }
  function isSpaceCombinator(node) {
    return node.type === "combinator" && /^\s+$/.test(node.value);
  }
  return {
    postcssPlugin: "vine-style-scoped",
    Rule(rule) {
      processRule(id, rule);
    },
    AtRule(node) {
      if (/-?keyframes$/.test(node.name) && !node.params.endsWith(`-${shortId}`)) {
        keyframes[node.params] = node.params = `${node.params}-${shortId}`;
      }
    },
    OnceExit(root) {
      if (Object.keys(keyframes).length) {
        root.walkDecls((decl) => {
          if (animationNameRE.test(decl.prop)) {
            decl.value = decl.value.split(",").map((v) => keyframes[v.trim()] || v.trim()).join(",");
          }
          if (animationRE.test(decl.prop)) {
            decl.value = decl.value.split(",").map((v) => {
              const vals = v.trim().split(/\s+/);
              const i = vals.findIndex((val) => keyframes[val]);
              if (i !== -1) {
                vals.splice(i, 1, keyframes[vals[i]]);
                return vals.join(" ");
              } else {
                return v;
              }
            }).join(",");
          }
        });
      }
    }
  };
};
scopedPlugin.postcss = true;

const trimPlugin = () => {
  return {
    postcssPlugin: "vine-style-trim",
    Once(root) {
      root.walk(({ type, raws }) => {
        if (type === "rule" || type === "atrule") {
          if (raws.before)
            raws.before = "\n";
          if ("after" in raws && raws.after)
            raws.after = "\n";
        }
      });
    }
  };
};
trimPlugin.postcss = true;

const cssVarsPlugin = (options) => {
  const { vineCompFns } = options.fileCtx;
  let cssBindings = {};
  vineCompFns.some((fnComp) => {
    if (fnComp.scopeId === options.scopeId) {
      cssBindings = fnComp.cssBindings;
      return true;
    }
    return false;
  });
  return {
    postcssPlugin: "vine-style-css-vars",
    Once(root) {
      root.walk((ctx) => {
        if (ctx.type === "decl" && ctx.value && ctx.value.includes("v-bind")) {
          for (const cbKey in cssBindings) {
            const range = {
              start: null,
              end: null
            };
            const content = parseCssVars(
              [ctx.value],
              {
                getIndex(start, end) {
                  range.start = start;
                  range.end = end;
                }
              }
            );
            if (content[0] === cbKey) {
              const mg = new MagicString(ctx.value);
              mg.overwrite(range.start, range.end, `--${cssBindings[cbKey]}`);
              mg.replaceAll("v-bind", "var");
              ctx.value = mg.toString();
              break;
            }
          }
        }
      });
    }
  };
};
cssVarsPlugin.postcss = true;

var __defProp$1 = Object.defineProperty;
var __defProps = Object.defineProperties;
var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols$1 = Object.getOwnPropertySymbols;
var __hasOwnProp$1 = Object.prototype.hasOwnProperty;
var __propIsEnum$1 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$1 = (obj, key, value) => key in obj ? __defProp$1(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$1 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$1.call(b, prop))
      __defNormalProp$1(a, prop, b[prop]);
  if (__getOwnPropSymbols$1)
    for (var prop of __getOwnPropSymbols$1(b)) {
      if (__propIsEnum$1.call(b, prop))
        __defNormalProp$1(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));
var __async = (__this, __arguments, generator) => {
  return new Promise((resolve, reject) => {
    var fulfilled = (value) => {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    };
    var rejected = (value) => {
      try {
        step(generator.throw(value));
      } catch (e) {
        reject(e);
      }
    };
    var step = (x) => x.done ? resolve(x.value) : Promise.resolve(x.value).then(fulfilled, rejected);
    step((generator = generator.apply(__this, __arguments)).next());
  });
};
function compileVineStyle(compilerCtx, params) {
  return __async(this, null, function* () {
    var _a, _b;
    let { source } = params;
    const {
      vineFileId,
      isScoped,
      scopeId,
      preprocessLang,
      inputSourceMap,
      preprocessOptions = (_a = compilerCtx.options.preprocessOptions) != null ? _a : {},
      postcssOptions = compilerCtx.options.postcssOptions,
      postcssPlugins: userPostcssPlugins = (_b = compilerCtx.options.postcssPlugins) != null ? _b : []
    } = params;
    const preprocessor = preprocessLang && processors[preprocessLang];
    const preProcessedSource = preprocessor == null ? void 0 : preprocessor(
      source,
      inputSourceMap,
      preprocessOptions
    );
    const map = preProcessedSource ? preProcessedSource.map : inputSourceMap;
    if (preProcessedSource) {
      source = preProcessedSource.code;
    }
    const postcssPlugins = [
      trimPlugin(),
      ...userPostcssPlugins
    ];
    if (isScoped) {
      const shortId = scopeId.replace(/^data-v-/, "");
      const longId = `data-v-${shortId}`;
      postcssPlugins.push(
        scopedPlugin({
          compilerCtx,
          id: longId,
          fileCtx: compilerCtx.fileCtxMap.get(vineFileId)
        })
      );
    }
    if (source.includes("v-bind")) {
      postcssPlugins.push(cssVarsPlugin({
        fileCtx: compilerCtx.fileCtxMap.get(vineFileId),
        scopeId
      }));
    }
    const postCSSOptions = __spreadProps(__spreadValues$1({}, postcssOptions), {
      from: void 0,
      to: void 0
    });
    if (map) {
      postCSSOptions.map = {
        inline: false,
        annotation: false,
        prev: map
      };
    }
    const errors = [];
    if (preProcessedSource && preProcessedSource.errors.length) {
      errors.push(...preProcessedSource.errors);
    }
    const dependencies = new Set(
      preProcessedSource ? preProcessedSource.dependencies : []
    );
    const recordPlainCssDependencies = (messages) => {
      messages.forEach((msg) => {
        if (msg.type === "dependency") {
          dependencies.add(msg.file);
        }
      });
      return dependencies;
    };
    const result = postcss(postcssPlugins).process(source, postCSSOptions);
    return result.then((result2) => ({
      code: result2.css || "",
      map: result2.map && result2.map.toJSON(),
      errors,
      rawResult: result2,
      dependencies: recordPlainCssDependencies(result2.messages)
    })).catch((error) => ({
      code: "",
      map: void 0,
      errors: [...errors, error],
      rawResult: void 0,
      dependencies
    }));
  });
}

var __defProp = Object.defineProperty;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
function createCompilerCtx(options) {
  return {
    fileCtxMap: /* @__PURE__ */ new Map(),
    vineCompileErrors: [],
    vineCompileWarnings: [],
    isRunningHMR: false,
    options: __spreadValues({
      inlineTemplate: true
    }, options)
  };
}
function createVineFileCtx(code, fileId, fileCtxCache) {
  const root = babelParse(code);
  const vineFileCtx = {
    root,
    fileId,
    renderOnly: fileCtxCache ? fileCtxCache.renderOnly : false,
    hmrCompFnsName: fileCtxCache ? fileCtxCache.hmrCompFnsName : null,
    fileSourceCode: new MagicString(code),
    vineCompFns: [],
    userImports: {},
    styleDefine: {},
    vueImportAliases: {},
    get originCode() {
      return this.fileSourceCode.original;
    }
  };
  return vineFileCtx;
}
function doValidateVine(vineCompilerHooks, vineFileCtx, vineCompFnDecls) {
  var _a;
  validateVine(vineCompilerHooks, vineFileCtx, vineCompFnDecls);
  (_a = vineCompilerHooks.onValidateEnd) == null ? void 0 : _a.call(vineCompilerHooks);
}
function doAnalyzeVine(vineCompilerHooks, vineFileCtx, vineCompFnDecls) {
  var _a;
  analyzeVine(vineCompilerHooks, vineFileCtx, vineCompFnDecls);
  (_a = vineCompilerHooks.onAnalysisEnd) == null ? void 0 : _a.call(vineCompilerHooks);
}
function compileVineTypeScriptFile(code, fileId, compilerHooks, fileCtxCache) {
  var _a, _b;
  const compilerOptions = compilerHooks.getCompilerCtx().options;
  const vineFileCtx = createVineFileCtx(code, fileId, fileCtxCache);
  (_a = compilerHooks.onBindFileCtx) == null ? void 0 : _a.call(compilerHooks, fileId, vineFileCtx);
  const vineCompFnDecls = findVineCompFnDecls(vineFileCtx.root);
  doValidateVine(compilerHooks, vineFileCtx, vineCompFnDecls);
  doAnalyzeVine(compilerHooks, vineFileCtx, vineCompFnDecls);
  transformFile(
    vineFileCtx,
    compilerHooks,
    (_b = compilerOptions == null ? void 0 : compilerOptions.inlineTemplate) != null ? _b : true
  );
  return vineFileCtx;
}

export { compileVineStyle, compileVineTypeScriptFile, createCompilerCtx, createVineFileCtx, doAnalyzeVine, doValidateVine, findVineCompFnDecls };
//# sourceMappingURL=index.mjs.map
